"""
There are currently three types of textures supported in blueptints.
Boxes, 2D-Planes and Meshtextures. To apply a texture to a Thing it has 
to be mediated through a :class:`Material <blueprints.material.Material>`.

.. code-block::
	:caption: Texture asssignment

	>>> texture  = blue.texture.Plane()
	>>> material = blue.Material(texture=texture)
	>>> plane    = blue.geoms.Plane(material=material)


Texture Types
=============
We will present the basic Texture Types here for image textures. The 
same types are also usable for Procedural Textures (see below).

Plane
-----

The :class:`Plane` texture encodes 2D images which can be provided by a file path. 
For clarity the MIME type and the number of channels can be made explicit with 
``content`` and ``n_channel``. To flip the image vertically or horizontally 
``v_flip`` and ``h_flip`` can be used.

.. code-block::
	:caption: Plane Texture

	>>> sand_tex = blue.texture.Plane(filename='sand.png')
	>>> sand_mat = blue.Material(texture=sand_tex)
	>>> blue.geoms.Plane(material=sand_mat)

.. image:: /_static/texture_plane.png

Box
---

There are two ways of specifying the six sides of a :class:`Box` we give the first 
as an example here and the second for the equivalent :class:`Skybox` texture.
The provided file paths are as always taken relative to the file instantiating the World.

.. code-block::
	:caption: Box Texture

	>>> grass_tex = blue.texture.Box(filename_up   ='grass_top.png', 
	>>> 				 filename_left ='grass_side.png', 
	>>> 				 filename_right='grass_side.png', 
	>>> 				 filename_front='grass_side.png', 
	>>> 				 filename_back ='grass_side.png', 
	>>> 				 filename_down ='grass_bottom.png')
	>>> grass_mat = blue.Material(name='grass', texture=grass_tex)
	>>> grass = blue.geoms.Box(material=grass_mat)

.. image:: /_static/texture_box.png


Mesh
----

If a :class:`Mesh <blueprints.geoms.Mesh>` has texture coordinates (like ``.obj`` files) 
a :class:'Plane' texture can be applied to it. If a mesh does not have texture coordinates 
they can also be provided manually from the user through :attr:`Mesh.texcoord <blueprints.assets.MeshAsset.texcoord>`.

.. code-block::
	:caption: Mesh Texture

	>> creeper_tex = blue.texture.Plane(filename='creeper.png')
	>> creeper_mat = blue.Material(texture=creeper_tex)
	>> creeper = blue.geoms.Mesh(filename='creeper.obj')

.. image:: /_static/texture_mesh.png

SkyBox
------

Instead of applying a :class:`Skybox` texture via a :class:`Material <blueprints.material.Material>` 
it can be applied directly to the :class:`World <blueprints.world.World>` through the :attr:`World.texture <blueprints.world.World.texture>` 
attribute. Skybox and Box textures use equvalent attributes and constructions.

.. warning::
	As of Mujoco 3.3.2 there is a bug that prevents a skybox from being loaded if no other textured :class:`Material <blueprints.material.Material>` 
	if included in the :class:`World <blueprints.world.World>`.

To load a cube like texture from a single file it is partitioned into a grid through the 
``grid_size`` argument. The number of cells (i.e. the product of ``grid_size``) must not 
exceed 12. The assignment of each cell to a side of the cube is then specified via the 
``grid_layout`` string. Each row is read left to right with the character,
``'U'`` up, ``'D'`` down, ``'L'`` left, ``'R'`` right, ``'F'`` front, ``'B'`` back and ``'.'`` skip 
mapping sides to grid cells.

.. image:: /_static/bluesky_grid.png

.. code-block::
	:caption: Skybox Texture

	>>> sky = blue.texture.Skybox(filename='sky.png', 
	>>> 			      grid_layout='.U..LFRB.D..', 
	>>> 			      grid_size=[3, 4])
	>>> world.texture = sky

.. image:: /_static/texture_sky.png


Procedural Textures
===================
Mujoco supports simple procedurally generated textures which are accessable 
through the bultin attribute. Their apperance for :class:`Box Textures <blueprints.texture.Box>` 
and :class:`Plane Textures <blueprints.texture.Plane>` differs so we present both in the 
following examples.

.. note::
	Though they are named identically and we present Box Geoms with Box Textures 
	and Plane geoms with Plane Textures, the opposite assignment is also possible.


.. code-block::
	:caption: Example

	>>> box_tex = ...
	>>> plane_tex = ...
	>>> ...
	>>> box_mat = blue.Material(texture=box_tex)
	>>> box = blue.geoms.Box(material=box_mat, 
	>>> 			 x=-1.5,
	>>> 			 z=0.5, 
	>>> 			 gamma=TAU/8)
	>>> plane_mat = blue.Material(texture=plane_tex, 
	>>> 			      texrepeat=[2, 1])
	>>> plane = blue.geoms.Plane(material=plane_mat, 
	>>> 			     x_length=2, 
	>>> 			     y_length=1,
	>>> 			     x=1.5, 
	>>> 			     z=0.5)
	>>> world.attach(box, plane, blue.geoms.Plane(color='orange'))


Gradient
--------

.. code-block::
	:caption: Gradient Builtin Texture

	>>> box_tex = blue.texture.Box(builtin='gradient', 
	>>> 			       color_1='white', 
	>>> 			       color_2='black', 
	>>> 			       width=1000, 
	>>> 			       height=1000)
	>>> plane_tex = blue.texture.Plane(builtin='gradient',
	>>> 				   color_1='white', 
	>>> 				   color_2='black', 
	>>> 				   width=1000, 
	>>> 				   height=1000)
	>>> ...


.. image:: /_static/texture_gradient.png


Checker
-------

.. code-block::
	:caption: Checker Builtin Texture

	>>> box_tex = blue.texture.Box(builtin='checker', 
	>>> 			       color_1='white', 
	>>> 			       color_2='black', 
	>>> 			       width=1000, 
	>>> 			       height=1000)
	>>> plane_tex = blue.texture.Plane(builtin='checker',
	>>> 				   color_1='white', 
	>>> 				   color_2='black', 
	>>> 				   width=1000, 
	>>> 				   height=1000)
	>>> ...


.. image:: /_static/texture_checker.png

Flat
----

.. code-block::
	:caption: Flat Builtin Texture

	>>> box_tex = blue.texture.Box(builtin='flat', 
	>>> 			       color_1='white', 
	>>> 			       color_2='black', 
	>>> 			       width=1000, 
	>>> 			       height=1000)
	>>> plane_tex = blue.texture.Plane(builtin='flat',
	>>> 				p   color_1='white', 
	>>> 				   color_2='black', 
	>>> 				   width=1000, 
	>>> 				   height=1000)
	>>> ...


.. image:: /_static/texture_flat.png


Mark
----

Additionally there are three variations of markings: ``'edge'``, ``'cross'`` and ``'random'``.
The last marking takes the additional argument ``random`` indicating the probability of a texture 
pixel being colored with the ``color_mark``.

.. code-block::
	:caption: Mark Texture

	>>> edge_tex = blue.texture.Box(builtin='flat', 
	>>> 				color_1='white', 
	>>> 				color_2='black', 
	>>> 				color_mark='orange', 
	>>> 				mark='edge', 
	>>> 				width=50, 
	>>> 				height=50)
	>>> cross_tex = blue.texture.Box(builtin='flat', 
	>>> 				 color_1='white', 
	>>> 				 color_2='black', 
	>>> 				 color_mark='orange', 
	>>> 				 mark='cross', 
	>>> 				 width=50, 
	>>> 				 height=50)
	>>> random_1_tex = blue.texture.Box(builtin='flat', 
	>>> 				    color_1='white', 
	>>> 				    color_2='black', 
	>>> 				    color_mark='orange', 
	>>> 				    mark='random', 
	>>> 				    random=0.05,
	>>> 				    width=50, 
	>>> 				    height=50)
	>>> random_2_tex = blue.texture.Box(builtin='flat', 
	>>> 				    color_1='white', 
	>>> 				    color_2='black', 
	>>> 				    color_mark='orange', 
	>>> 				    mark='random', 
	>>> 				    random=0.5,
	>>> 				    width=50, 
	>>> 				    height=50)


.. image:: /_static/texture_mark.png

"""


import numpy as np
import blueprints as blue
from blueprints.thing.colored import Color



class BaseTexture(blue.BaseThing):
	"""
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#asset-texture>`__.
	"""
	@blue.restrict
	def _build(self, 
		   parent, 
		   world, 
		   indicies, 
		   **kwargs):
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Thing.
		"""
		self.asset._build(parent=parent, 
				  world=world, 
				  indicies=indicies, 
				  **kwargs)
		return self.asset._xml_root


	@property
	def asset(self) -> blue.TextureAssetType:
		"""
		The attributes of the texture are stored in a separate asset object.
		If this texture is assign to multiple objects, and then gets modified 
		through ``Material.texture.attribute = value`` all  other Materials 
		referenceing the Texture will be unaltered. See :class:`UniqueThing <blueprints.thing.unique.UniqueThing>`.

		Returns
		-------
		blue.TextureAssetType
		"""
		return self._asset


	@asset.setter
	@blue.restrict
	def asset(self, asset: blue.TextureAssetType) -> None:
		self._asset = asset	

	# BLUEPRINT PROPERTIES

	@property
	def content(self) -> str|None:
		"""
		Content type (MIME type) taking values ``None``, ``'image/png'``,
		``'image/ktx'`` and ``'image/vnd.mujoco.texture'``.

		Returns
		-------
		str | None
		"""
		return self.asset.content


	@content.setter
	@blue.restrict
	def content(self, content: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.content = content


	@property
	def filename(self) -> str|None:
		"""
		Name of the texture file.

		Returns
		-------
		str | None
		"""
		return self.asset.file


	@filename.setter
	@blue.restrict
	def filename(self, filename: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.file = filename


	@property
	def grid_size(self) -> list[int]:
		"""
		Grid size attrigute ``[n_rows, n_cols]`` defining the grid. 
		For non-grid Textures this value is ignored.

		Returns
		-------
		list[int]
		"""
		return self.asset.grid_size


	@grid_size.setter
	@blue.restrict
	def grid_size(self, grid_size: list[int]) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.grid_size = grid_size


	@property
	def grid_layout(self) -> str|None:
		"""
		The layout of the grid. See example above.

		Returns
		-------
		str | None
		"""
		return self.asset.grid_layout


	@grid_layout.setter
	@blue.restrict
	def grid_layout(self, grid_layout: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.grid_layout = grid_layout


	@property
	def filename_right(self) -> str|None:
		"""
		The file name for the right side of a :class:`Box`.

		Returns
		-------
		str | None
		"""
		return self.asset.filename_right


	@filename_right.setter
	@blue.restrict
	def filename_right(self, filename_right: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.filename_right = filename_right


	@property
	def filename_left(self) -> str|None:
		"""
		The file name for the left side of a :class:`Box`.
		
		Returns
		-------
		str | None
		"""
		return self.asset.filename_left


	@filename_left.setter
	@blue.restrict
	def filename_left(self, filename_left: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.filename_left = filename_left


	@property
	def filename_up(self) -> str|None:
		"""
		The file name for the upper side of a :class:`Box`.
		
		Returns
		-------
		str | None
		"""
		return self.asset.filename_up


	@filename_up.setter
	@blue.restrict
	def filename_up(self, filename_up: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.filename_up = filename_up


	@property
	def filename_down(self) -> str|None:
		"""
		The file name for the down side of a :class:`Box`.
		
		Returns
		-------
		str | None
		"""
		return self.asset.filename_down


	@filename_down.setter
	@blue.restrict
	def filename_down(self, filename_down: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.filename_down = filename_down


	@property
	def filename_front(self) -> str|None:
		"""
		The file name for the front side of a :class:`Box`.
		
		Returns
		-------
		str | None
		"""
		return self.asset.filename_front


	@filename_front.setter
	@blue.restrict
	def filename_front(self, filename_front: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.filename_front = filename_front


	@property
	def filename_back(self) -> str|None:
		"""
		The file name for the back side of a :class:`Box`.
		
		Returns
		-------
		str | None
		"""
		return self.asset.filename_back


	@filename_back.setter
	@blue.restrict
	def filename_back(self, filename_back: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.filename_back = filename_back


	@property
	def builtin(self) -> str|None:
		"""
		If set, this attribute specifies a builtin procedural Texture. 
		See examples above.

		Returns
		-------
		str | None
		"""
		return self.asset.builtin


	@builtin.setter
	@blue.restrict
	def builtin(self, builtin: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.builtin = builtin


	@property
	def color_1(self) -> blue.ColorType:
		"""
		The first color of a procedural Texture.

		Returns
		-------
		blue.ColorType
		"""
		return self.asset.color_1


	@color_1.setter
	@blue.restrict
	def color_1(self, color_1: object) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.color_1 = color_1


	@property
	def color_2(self) -> blue.ColorType:
		"""
		The second color of a procedural Texture.
		
		Returns
		-------
		blue.ColorType
		"""
		return self.asset.color_2


	@color_2.setter
	@blue.restrict
	def color_2(self, color_2: object) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.color_2 = color_2


	@property
	def mark(self) -> str|None:
		"""
		If set this attribute specifies additional markings on 
		a procedural Texture. Possible values are ``None``, 
		``'edge'``, ``'cross'`` and ``'random'``. See examples above.
		
		Returns
		-------
		str | None
		"""
		return self.asset.mark


	@mark.setter
	@blue.restrict
	def mark(self, mark: str|None) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.mark = mark


	@property
	def color_mark(self) -> blue.ColorType:
		"""
		The mark color of a procedural Texture.
		
		Returns
		-------
		blue.ColorType
		"""
		return self.asset.color_mark


	@color_mark.setter
	@blue.restrict
	def color_mark(self, color_mark: object) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.color_mark = color_mark


	@property
	def random(self) -> float:
		"""
		If the ``mark`` attribute is set to ``'random'``, this 
		attribute specifies the probability of a pixel in the 
		texture being the ``mark_color``.

		Returns
		-------
		float
		"""
		return self.asset.random


	@random.setter
	@blue.restrict
	def random(self, random: float) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.random = random


	@property
	def width(self) -> int:
		"""
		The width of a procedural texture, i.e., the number of columns 
		in the image. Larger values usually result in higher quality 
		images, although in some cases (e.g. checker patterns) small 
		values are sufficient. For textures loaded from files, this 
		attribute is ignored.

		Returns
		-------
		int
		"""
		return self.asset.width


	@width.setter
	@blue.restrict
	def width(self, width: int) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.width = width


	@property
	def height(self) -> int:
		"""
		The height of the procedural texture, i.e., the number of rows 
		in the image. For :class:`Box` and :class:`Skybox` textures, this attribute is 
		ignored and the height is set to 6 times the width. For textures 
		loaded from files, this attribute is ignored.

		Returns
		-------
		int
		"""
		return self.asset.height


	@height.setter
	@blue.restrict
	def height(self, height: int) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.height = height


	@property
	def h_flip(self) -> bool:
		"""
		If true, images loaded from file are flipped in the horizontal 
		direction. Does not affect procedural textures.

		Returns
		-------
		bool
		"""
		return self.asset.h_flip


	@h_flip.setter
	@blue.restrict
	def h_flip(self, h_flip: bool) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.h_flip = h_flip


	@property
	def v_flip(self) -> bool:
		"""
		If true, images loaded from file are flipped in the vertical 
		direction. Does not affect procedural textures.

		Returns
		-------
		bool
		"""
		return self.asset.v_flip


	@v_flip.setter
	@blue.restrict
	def v_flip(self, v_flip: bool) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.v_flip = v_flip


	@property
	def n_channel(self) -> int:
		"""
		The number of channels in the texture image file. This allows 
		loading 4-channel textures (RGBA) or single-channel textures 
		(e.g., for Physics-Based Rendering properties such as roughness 
		or metallic).

		Returns
		-------
		int
		"""
		return self.asset.n_channel


	@n_channel.setter
	@blue.restrict
	def n_channel(self, n_channel: int) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.n_channel = n_channel



class Plane(blue.PlaneTextureType, BaseTexture):
	@blue.restrict
	def __init__(self, 
		     content:     str|None  =  None, 
		     filename:    str|None  =  None, 
		     builtin:     str|None  =  None, 
		     color_1:     object    = [0.8, 0.8, 0.8], 
		     color_2:     object    = [0.5, 0.5, 0.5], 
		     mark:        str|None  =  None, 
		     color_mark:  object    = [0.0, 0.0, 0.0],
		     random:      float     =  0.01, 
		     width:       int       =  1000, 
		     height:      int       =  1000, 
		     h_flip:      bool      =  False, 
		     v_flip:      bool      =  False, 
		     n_channel:   int       =  3, 
		     name:        str|None  = None, 
		     asset:       blue.TextureAssetType|None = None):
		if asset is None:
			self._asset = blue.assets.TextureAsset(content=content, 
							       file=filename, 
							       builtin=builtin, 
							       color_1=color_1, 
							       color_2=color_2, 
							       mark=mark, 
							       color_mark=color_mark, 
							       random=random, 
							       width=width, 
							       height=height, 
							       h_flip=h_flip, 
							       v_flip=v_flip, 
							       n_channel=n_channel, 
							       name=name)
		else:
			self._asset = asset
		self.asset._add(self)
		self.asset._TYPE = self._TYPE
		super().__init__(name=name)



class Box(blue.BoxTextureType, BaseTexture):
	@blue.restrict
	def __init__(self, 
		     content:         str|None  =  None, 
		     filename:        str|None  =  None, 
		     grid_size:       list[int] = [1, 1], 
		     grid_layout:     str|None  =  None,
		     filename_right:  str|None  =  None, 
		     filename_left:   str|None  =  None, 
		     filename_up:     str|None  =  None, 
		     filename_down:   str|None  =  None, 
		     filename_front:  str|None  =  None, 
		     filename_back:   str|None  =  None, 
		     builtin:         str|None  =  None, 
		     color_1:         object    = [0.8, 0.8, 0.8], 
		     color_2:         object    = [0.5, 0.5, 0.5], 
		     mark:            str|None  =  None, 
		     color_mark:      object    = [0.0, 0.0, 0.0],
		     random:          float     =  0.01, 
		     width:           int       =  1000, 
		     height:          int       =  1000, 
		     h_flip:          bool      =  False, 
		     v_flip:          bool      =  False, 
		     n_channel:       int       =  3, 
		     name:            str|None  = None, 
		     asset:           blue.TextureAssetType|None = None):
		if asset is None:
			self._asset = blue.assets.TextureAsset(content=content, 
							       file=filename, 
							       grid_size=grid_size, 
							       grid_layout=grid_layout, 
							       fileright=filename_right, 
							       fileleft=filename_left, 
							       fileup=filename_up, 
							       filedown=filename_down, 
							       filefront=filename_front, 
							       fileback=filename_back, 
							       builtin=builtin, 
							       color_1=color_1, 
							       color_2=color_2, 
							       mark=mark, 
							       color_mark=color_mark, 
							       random=random, 
							       width=width, 
							       height=height, 
							       h_flip=h_flip, 
							       v_flip=v_flip, 
							       n_channel=n_channel, 
							       name=name)
		else:
			self._asset = asset
		self.asset._add(self)
		self.asset._TYPE = self._TYPE
		super().__init__(name=name)



class Skybox(blue.SkyboxTextureType, Box):
	@blue.restrict
	def __init__(self, 
		     content:         str|None  =  None, 
		     filename:        str|None  =  None, 
		     grid_size:       list[int] = [1, 1], 
		     grid_layout:     str|None  =  None,
		     filename_right:  str|None  =  None, 
		     filename_left:   str|None  =  None, 
		     filename_up:     str|None  =  None, 
		     filename_down:   str|None  =  None, 
		     filename_front:  str|None  =  None, 
		     filename_back:   str|None  =  None, 
		     builtin:         str|None  =  None, 
		     color_1:         object    = [0.8, 0.8, 0.8], 
		     color_2:         object    = [0.5, 0.5, 0.5], 
		     mark:            str|None  =  None, 
		     color_mark:      object    = [0.0, 0.0, 0.0],
		     random:          float     =  0.01, 
		     width:           int       =  1000, 
		     height:          int       =  1000, 
		     h_flip:          bool      =  False, 
		     v_flip:          bool      =  False, 
		     n_channel:       int       =  3, 
		     name:            str|None  = None, 
		     asset:           blue.TextureAssetType|None = None):
		super().__init__(content=content, 
				 filename=filename, 
				 grid_size=grid_size, 
				 grid_layout=grid_layout, 
				 filename_right=filename_right, 
				 filename_left=filename_left, 
				 filename_up=filename_up, 
				 filename_down=filename_down, 
				 filename_front=filename_front, 
				 filename_back=filename_back, 
				 builtin=builtin, 
				 color_1=color_1, 
				 color_2=color_2, 
				 mark=mark, 
				 color_mark=color_mark, 
				 random=random, 
				 width=width, 
				 height=height, 
				 h_flip=h_flip, 
				 v_flip=v_flip, 
				 n_channel=n_channel, 
				 name=name, 
				 asset=asset)



import blueprints as blue
from blueprints import restrict
from . import base

import numpy as np
import inspect
import xml.etree.ElementTree as xml
from copy import copy
from collections import defaultdict



@blue.restrict
def gradient(*colors: list[str|int|list[int|float]|np.ndarray|blue.ColorType], 
	     n_steps: int):
	"""
	This function creates a color gradient in form of a list of colors, 
	which are interpolated between argument color.

	>>> blue.gradient('red', 'yellow', n_steps=6)
	[Red[#FF0000], Red[#F52900], Orange[#EB5200], Orange[#E07A00], Yellow[#D6A300], Yellow[#CCCC00]]


	Parameters
	----------
	colors : list
		A list of colors given in any of the formats specified by :class:`Color`.
	n_steps : int
		The number of steps taken from the first to the last color.

	Returns
	-------
	list
		List of colors interpolated in n_steps along all given colors.
	"""
	n_colors = len(colors)
	if n_colors < 2:
		raise ValueError(f'At least two colors must be given. Got {n_colors} instead.')
	if n_steps < 1:
		raise ValueError(f'Argument n_steps must be larger than 0, got {n_steps} instead.')
	if n_colors > n_steps:
		raise ValueErrur(f'There must be more intermediate steps then colors')
	ratio = (n_colors-1)/(n_steps - 1)
	func  = lambda i, j: max(0, 1 - abs(i * ratio - j))
	rate  = np.array([[[func(i, j)] for j in range(n_colors)]\
					for i in range(n_steps)])
	rgbs  = np.array([Color(color).rgb for color in colors])[None,...]
	grad  = np.sum(rgbs * rate, axis=1)
	grad  = np.maximum(0, np.minimum(1, grad))
	return list(map(Color, np.rollaxis(grad, axis=0)))



class Color(blue.ColorType):

	"""
	The color descriptor can be one of the following:

	 * string containing the colors name: ``'orange'``
	 * string containing the colors hex code as RGB: ``'#FFFF00'``
	 * string containing the colors hex code as RGBA: ``'#FFFF0080'``
	 * integer specifying the colors hex code as RGB ``0xFFFF00``
	 * list containing the colors components as RGB: ``[1, 1, 0]``
	 * list containing the colors components as RGBA: ``[1, 1, 0, 0.5]``
	 * np.array containing the colors components as RGB: ``[1, 1, 0]``
	 * np.array containing the colors components as RGBA: ``[1, 1, 0, 0.5]``
	 * Color object.
	"""

	@restrict
	def __init__(self, 
		     descriptor: str|int|list[int|float]|np.ndarray|blue.ColorType, 
		     opacity: int|float|None = None):
		if isinstance(opacity, (int, float)) and not 0 <= opacity <= 1:
			raise ValueError(f'Opacity must either be None or in range [0, 1], got {opacity} instead.')
		try:
			if isinstance(descriptor, str):
				rgba = self._from_string(descriptor, opacity)
			elif isinstance(descriptor, int):
				rgba = self._from_int(descriptor, opacity)
			elif isinstance(descriptor, list):
				rgba = self._from_list(descriptor, opacity)
			elif isinstance(descriptor, np.ndarray):
				rgba = self._from_array(descriptor, opacity)
			elif isinstance(descriptor, blue.ColorType):
				rgba = descriptor.rgba
			else:
				raise Exception('This line should never be reached. Please reboot your local universe.')
		except ValueError as error:
			raise error from None
		self.rgba = rgba


	def __repr__(self):
		"""
		The representations contains the name of the color and the RGB hex value.

		Returns
		-------
		str
		"""
		return f'{self.name.title()}[#{self._hex_from_array(self.rgb).upper()}]'


	@property
	def name(self) -> str:
		"""
		The name is set to be the closest name among default colors.

		Returns
		-------
		str
		"""
		return min(self._COLORS.items(), key=lambda x: np.linalg.norm(self.rgb - x[1]))[0]


	@property
	def rgb(self) -> np.ndarray:
		"""
		The values are synchronized with :attr:`rgba`, :attr:`red`,:attr:`green` and :attr:`blue` .

		Returns
		-------
		np.ndarray
			The color array without opacity.
		"""
		return self._rgba[:3]


	@rgb.setter
	@restrict
	def rgb(self, rgb: list[int|float]|np.ndarray) -> None:
		"""
		The values are synchronized with :attr:`rgba`, :attr:`red`,:attr:`green` and :attr:`blue` .

		Parameters
		----------
		rgb : list[int | float] | np.ndarray
			The color array without opacity.

		Raises
		------
		ValueError
			If any component is not in the range [0, 1] an error is raised.
		"""
		rgb = np.array(rgb, dtype=np.float32)
		if not np.all(0 <= rgb) and np.all(rgb <= 1):
			raise ValueError(f'All color components must be in range [0, 1], got {rgb}.')
		self._rgba[:3] = rgb


	@property
	def rgba(self) -> np.ndarray:
		"""
		The values are synchronized with :attr:`rgb`, :attr:`red`,:attr:`green`, :attr:`blue` and :attr:`opacity`.

		Returns
		-------
		np.ndarray
			The color array with opacity.
		"""
		return self._rgba


	@rgba.setter
	@restrict
	def rgba(self, rgba: list[int|float]|np.ndarray) -> None:
		"""
		The values are synchronized with :attr:`rgb`, :attr:`red`,:attr:`green`, :attr:`blue` and :attr:`opacity`.

		Parameters
		----------
		rgba : list[int | float] | np.ndarray
			The color array with opacity.

		Raises
		------
		ValueError
			If any component is not in the range [0, 1] an error is raised.
		"""
		rgba = np.array(rgba, dtype=np.float32)
		if not np.all(0 <= rgba) and np.all(rgba <= 1):
			raise ValueError(f'All color components must be in range [0, 1], got {rgba}.')
		self._rgba = rgba


	@property
	def red(self) -> float|None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.

		Returns
		-------
		float | None
			The red component of the color array.
		"""
		return float(self._rgba[0])


	@red.setter
	@restrict
	def red(self, red: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.

		Parameters
		----------
		red : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The red component of the color array.

		Raises
		------
		ValueError
			If the red component is not in the range [0, 1] an error is raised.
		"""
		if not 0 <= red <= 1:
			raise ValueError(f'All color components must be in range [0, 1], got {red}.')
		self._rgba[0] = float(red)


	@property
	def green(self) -> float|None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.
		
		Returns
		-------
		float | None
			The green component of the color array.
		"""
		return float(self._rgba[1])


	@green.setter
	@restrict
	def green(self, green: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.
		
		Parameters
		----------
		green : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The green component of the color array.

		Raises
		------
		ValueError
			If the green component is not in the range [0, 1] an error is raised.
		"""
		if not 0 <= green <= 1:
			raise ValueError(f'All color components must be in range [0, 1], got {green}.')
		self._rgba[1] = float(green)


	@property
	def blue(self) -> float|None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.
		
		Returns
		-------
		float | None
			The blue component of the color array.
		"""
		return float(self._rgba[2])


	@blue.setter
	@restrict
	def blue(self, blue: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.
		
		Parameters
		----------
		blue : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The blue component of the color array.

		Raises
		------
		ValueError
			If the blue component is not in the range [0, 1] an error is raised.
		"""
		if not 0 <= blue <= 1:
			raise ValueError(f'All color components must be in range [0, 1], got {blue}.')
		self._rgba[2] = float(blue)


	@property
	def opacity(self) -> int|float:
		"""
		The value is synchronized with :attr:`rgba`.
		
		Returns
		-------
		int | float
			The opacity component of the color array.
		"""
		return self._rgba[3]


	@opacity.setter
	@restrict
	def opacity(self, opacity: int|float) -> None:
		"""
		The value is synchronized with :attr:`rgba`.
		
		Parameters
		----------
		opacity : int | float
			The opacity component of the color array.

		Raises
		------
		ValueError
			If the opacity component is not in the range [0, 1] an error is raised.
		"""
		if not 0 <= opacity <= 1:
			raise ValueError(f'Opacity must be in range [0, 1], got {opacity}.')
		self._rgba[3] = float(opacity)



	# INIT METHODS

	@classmethod
	def _from_string(cls, descriptor, opacity):
		if descriptor.startswith('#'):
			rgba = cls._from_hex_string(descriptor, opacity)
		else:
			rgba = cls._from_name_string(descriptor, opacity)
		return rgba


	@classmethod
	def _from_int(cls, descriptor, opacity):
		red, green, blue = cls._rgb_from_hex(descriptor)
		opacity = opacity if opacity is not None else 1.0
		return [red/0xff, green/0xff, blue/0xff, opacity]


	@classmethod
	def _from_list(cls, descriptor, opacity):
		if len(descriptor) == 3:
			opacity = opacity if opacity is not None else 1.0
			rgba = descriptor + [opacity]
		elif len(descriptor) == 4:
			if opacity is None:
				rgba = descriptor
			else:
				opacity = opacity if opacity is not None else 1.0
				rgba = descriptor[:3] + [opacity]
		else:
			raise ValueError(f'The allowed lengths for descriptor as list is either 3 or 4, got {len(list)} instead.')
		return rgba


	@classmethod
	def _from_array(cls, descriptor, opacity):
		if descriptor.shape == (3,):
			opacity = opacity if opacity is not None else 1.0
			rgba = list(map(float, descriptor)) + [opacity]
		elif descriptor.shape == (4,):
			if opacity is None:
				rgba = list(map(float, descriptor))
			else:
				opacity = opacity if opacity is not None else 1.0
				rgba = list(map(float, descriptor[:3])) + [opacity]
		return rgba


	@classmethod
	def _from_hex_string(cls, descriptor, opacity):
		error = ValueError(f"""The string descriptor of a rgb or rgba hex code is not formated correctly. The received argument was {descriptor}.

RGB may be formated like this: 
'#E73D8A'
'#e73d8a'
RGBA may be formated like this:
'#E73D8AFF'
'#e73d8aff'
""")
		if len(descriptor) != 7 and len(descriptor) != 9:
			raise error
		NUMERALS = '0123456789abcdefABCDEF'
		if any(c not in NUMERALS for c in descriptor[1:]):
			raise error
		if len(descriptor) == 7:
			red, green, blue = cls._rgb_from_hex(int(descriptor[1:], base=16))
			opacity = opacity if opacity is not None else 1.0
			rgba = [red/0xff, green/0xff, blue/0xff, opacity]
		elif len(descriptor) == 9:
			red, green, blue, opacity = cls._rgba_from_hex(int(descriptor[1:], base=16))
			rgba = [red/0xff, green/0xff, blue/0xff, opacity/0xff]
		return rgba

	@classmethod
	def _from_name_string(cls, descriptor, opacity):
		if descriptor not in cls._COLORS:
			raise ValueError(f"Color expected descriptor to be {', '.join(cls._COLORS.keys())}, got {descriptor} instead.")
		opacity = opacity if opacity is not None else 1.0
		return cls._COLORS[descriptor] + [opacity]

	# CONVERSION METHODS

	@classmethod
	def _hex_from_array(cls, array):
		return ''.join(map(lambda n: hex(int(n))[2:].zfill(2), np.round(array * 0xff)))


	@classmethod
	def _rgb_from_hex(cls, n):
		"""
		Helper function to get the rgb from n

		Parameters
		----------
		n : int
			The int gets interpreted as rgb in hex

		Returns
		-------
		int, int, int
		"""
		red, green, blue = map(lambda p: cls._from_hex_partition(n, p), [2, 1, 0])
		return red, green, blue


	@classmethod
	def _rgba_from_hex(cls, n):
		"""
		Helper function to get the rgba from n

		Parameters
		----------
		n : int
			The int gets interpreted as rgba in hex

		Returns
		-------
		int, int, int, int
		"""
		red, green, blue, opacity = map(lambda p: cls._from_hex_partition(n, p), [3, 2, 1, 0])
		return red, green, blue, opacity


	@staticmethod
	def _from_hex_partition(n, p) -> int:
		"""
		Helper function to get the p-th partition of the hex int (index is read from right to left).

		Returns
		-------
		int
		"""
		return (n >> (p * 8)) & 0xff



class ColoredThing(blue.ColoredThingType, base.BaseThing):

	"""
	This class is used to implement color. The color components are stored in the :attr:`rgba` property. 
	Those components are also accessible as individual properties :attr:`red`, :attr:`green`, :attr:`blue` 
	and :attr:`opacity`. All components of the color array must be in the range [0, 1].

	.. note::
		The transparency of :class:`ColoredThing` is modified via :attr:`opacity`, and not with 
		``alpha``, since this property is reserved for :attr:`blueprints.thing.moveable.MoveableThing.alpha`!
	
	Parameters
	----------
	rgb : list[int | float] | np.ndarray | None, optional
		The color array without opacity.
	rgba : list[int | float] | np.ndarray | None, optional
		The color array with opacity.
	**kwargs
		Keyward arguments are passed to ``super().__init__``.
	"""

	@restrict
	def __init__(self, 
		     color='white', 
		     opacity=None, 
		     **kwargs):
		"""
		Parameters
		----------
		rgb : list[int | float] | np.ndarray | None, optional
			The color array without opacity.
		rgba : list[int | float] | np.ndarray | None, optional
			The color array with opacity.
		**kwargs
			Keyward arguments are passed to ``super().__init__``.
		"""
		self.color = color
		if opacity:
			self.opacity = opacity
		super().__init__(**kwargs)


	@property
	def color(self) -> blue.ColorType:
		"""
		The :class:`Color` contains all attributes of the :class:`ColoredThing` color components.

		Returns
		-------
		Color
		"""
		return self._color


	@color.setter
	@restrict
	def color(self, color: str|int|list[int|float]|np.ndarray|blue.ColorType|None) -> None:
		"""
		The color attribute is of the type:class:`Color` which contains all attributes of the 
		:class:`ColoredThing` color components.

		Parameters
		----------
		color : str|int|list[int|float]|np.ndarray|blue.ColorType
			The color descriptor can be one of the following:

			 * A string containing the colors name, for example ``'orange'``
			 * A string containing the colors hex code as RGB, for example ``'#FFFF00'``
			 * A string containing the colors hex code as RGBA, for example ``'#FFFF0080'``
			 * A list containing the colors components as RGB, for example ``[1, 1, 0]``
			 * A list containing the colors components as RGBA, for example ``[1, 1, 0, 0.5]``
			 * A np.array containing the colors components as RGB, for example ``[1, 1, 0]``
			 * A np.array containing the colors components as RGB, for example ``[1, 1, 0, 0.5]``
			 * A Color object.
		"""
		if color is not None:
			self._color = Color(color)
		else:
			self._color = None



	@property
	def rgb(self) -> np.ndarray:
		"""
		The values are synchronized with :attr:`rgba`, :attr:`red`,:attr:`green` and :attr:`blue` .

		Returns
		-------
		np.ndarray
			The color array without opacity.
		"""
		return self.color.rgb


	@rgb.setter
	@restrict
	def rgb(self, rgb: list[int|float]|np.ndarray) -> None:
		"""
		The values are synchronized with :attr:`rgba`, :attr:`red`,:attr:`green` and :attr:`blue` .

		Parameters
		----------
		rgb : list[int | float] | np.ndarray
			The color array without opacity.

		Raises
		------
		ValueError
			If any component is not in the range [0, 1] an error is raised.
		"""
		self.color.rgb = rgb


	@property
	def rgba(self) -> np.ndarray|None:
		"""
		The values are synchronized with :attr:`rgb`, :attr:`red`,:attr:`green`, :attr:`blue` and :attr:`opacity`.

		Returns
		-------
		np.ndarray
			The color array with opacity.
		"""
		if self.color is not None:
			return self.color.rgba
		else:
			return None


	@rgba.setter
	@restrict
	def rgba(self, rgba: list[int|float]|np.ndarray) -> None:
		"""
		The values are synchronized with :attr:`rgb`, :attr:`red`,:attr:`green`, :attr:`blue` and :attr:`opacity`.

		Parameters
		----------
		rgba : list[int | float] | np.ndarray
			The color array with opacity.

		Raises
		------
		ValueError
			If any component is not in the range [0, 1] an error is raised.
		"""
		if self.color is None:
			self.color = Color([0, 0, 0])
		self.color.rgba = rgba


	@property
	def red(self) -> float|None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.

		Returns
		-------
		float | None
			The red component of the color array.
		"""
		return self.color.red


	@red.setter
	@restrict
	def red(self, red: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.

		Parameters
		----------
		red : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The red component of the color array.

		Raises
		------
		ValueError
			If the red component is not in the range [0, 1] an error is raised.
		"""
		if self.color is None:
			self.color = Color([0, 0, 0])
		self.color.red = red


	@property
	def green(self) -> float|None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.
		
		Returns
		-------
		float | None
			The green component of the color array.
		"""
		return self.color.green


	@green.setter
	@restrict
	def green(self, green: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.
		
		Parameters
		----------
		green : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The green component of the color array.

		Raises
		------
		ValueError
			If the green component is not in the range [0, 1] an error is raised.
		"""
		if self.color is None:
			self.color = Color([0, 0, 0])
		self.color.green = green


	@property
	def blue(self) -> float|None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.
		
		Returns
		-------
		float | None
			The blue component of the color array.
		"""
		return self.color.blue


	@blue.setter
	@restrict
	def blue(self, blue: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		"""
		The value is synchronized with :attr:`rgb` and :attr:`rgba`.
		
		Parameters
		----------
		blue : int | float | np.int32 | np.int64 | np.float32 | np.float64
			The blue component of the color array.

		Raises
		------
		ValueError
			If the blue component is not in the range [0, 1] an error is raised.
		"""
		if self.color is None:
			self.color = Color([0, 0, 0])
		self.color.blue = blue


	@property
	def opacity(self) -> int|float:
		"""
		The value is synchronized with :attr:`rgba`.
		
		Returns
		-------
		int | float
			The opacity component of the color array.
		"""
		return self.color.opacity


	@opacity.setter
	@restrict
	def opacity(self, opacity: int|float) -> None:
		"""
		The value is synchronized with :attr:`rgba`.
		
		Parameters
		----------
		opacity : int | float
			The opacity component of the color array.

		Raises
		------
		ValueError
			If the opacity component is not in the range [0, 1] an error is raised.
		"""
		if self.color is None:
			self.color = Color([0, 0, 0])
		self.color.opacity = opacity



"""
If the children attributes of a :class:`NodeThing <blueprints.thing.node.NodeThing>` are called, those children will be 
returned in a :class:`View`. The :class:`View` can be used to get and set attributes of multiple objects 
as well as to call methods on those child Things simultaneously. The following example shows how to build 
a simplified hand model using Views.

.. code-block:: python
	:caption: Getting attributes
		
	>>> names = ['thumb', 'index', 'middle', 'ring', 'pinky']
	>>> fingers = [blue.Body(name=name) for name in names]
	>>> palm = blue.Body(name='palm', bodies=fingers)
	>>> view = palm.bodies # calling child attributes creates a view
	>>> view
	View[5|palm.bodies]
	>>> view.name
	['thumb', 'index', 'middle', 'ring', 'pinky']
	>>> view.bodies
	View[0|palm.bodies.bodies]
	>>> intermediate = blue.Body(name='intermediate')
	>>> distal = blue.Body(name='distal')
	>>> phalanges = [distal, intermediate, intermediate, intermediate, intermediate]
	>>> view.attach.distribute(phalanges) 
	>>> view.bodies[1:].attach(blue.Body(name='distal'))

.. code-block:: python
	:caption: Setting attributes
		
	>>> view.bodies[0].name = 'distal_thumb'
	>>> view.bodies[1:].name = ['intermediate_index', 
				    'intermediate_middle', 
				    'intermediate_ring', 
				    'intermediate_pinky']
	>>> view.bodies.name
	['distal_thumb', 
	 'intermediate_index', 
	 'intermediate_middle',
	 'intermediate_ring', 
	 'intermediate_pinky']

.. code-block::
	:caption: Applying functions
		
	>>> renaming = lambda x: x.__setattr__('name', f'distal_{x.parent.parent.name}')
	>>> view.bodies.bodies.apply(renaming)
	>>> view.bodies.bodies.name
	['distal_index', 'distal_middle', 'distal_ring', 'distal_pinky']

.. code-block:: mxml
	:caption: XML structure

	<body name="palm">
	    <body name="thumb">
	        <body name="distal_thumb" />
	    </body>
	    <body name="index">
	        <body name="intermediate_index">
	            <body name="distal_index" />
	        </body>
	    </body>
	    <body name="middle">
	        <body name="intermediate_middle">
	            <body name="distal_middle" />
	        </body>
	    </body>
	    <body name="ring">
	        <body name="intermediate_ring">
	            <body name="distal_ring" />
	        </body>
	    </body>
	    <body name="pinky">
	        <body name="intermediate_pinky">
	            <body name="distal_pinky" />
	        </body>
	    </body>
	</body>
"""

import numpy as np
import blueprints as blue
from collections import defaultdict, deque
from itertools import count, chain
from types import BuiltinFunctionType, FunctionType, MethodType



class View(blue.ViewType):
	"""
	The View class is used to handle the access to children of nodes in a kinematic tree and their properties.
	"""
	__RESERVED = {'_View__ELEMENTS', 
		      '_View__NAME', 
		      '_View__PARENT', 
		      '_View__RESERVED', 
		      '_AllView__ELEMENTS', 
		      '_AllView__NAME', 
		      '_AllView__PARENT', 
		      '_AllView__RESERVED', 
		      '_AllView__DESCENDANTS'}
	
	@blue.restrict
	def __init__(self, elements: list, name: str, parent: blue.ThingType|blue.LatticeType) -> None:
		"""
		The View can be used to get and set attributes of multiple objects at the same time.
		
		Parameters
		----------
		elements : list
			All Things, that are part of the View.
		name : str
			The name of the attribute that was requested for this View.
		parent : blue.ThingType
			The parent of the View.
		"""
		# ELIMINATE DUPLICATES (order-preserving, O(n))
		elements = list(dict.fromkeys(elements))
		self.__ELEMENTS = elements.copy()
		self.__NAME     = name
		self.__PARENT   = parent


	def __repr__(self) -> str:
		"""
		Represents the View as a string.
		
		Returns
		-------
		str
			The representing string contains a representation of the call that was used to get the View.
		"""
		return f'View[{len(self)}|{self.__PARENT.name}.{self.__NAME}]'


	def __len__(self):
		"""
		The number of Things contained in the View.
		
		Returns
		-------
		int
			The number of Things contained in the View.
		"""
		return len(self.__ELEMENTS)


	def __iter__(self):
		"""
		Used to iterate over all Things in the View. While iterating over the View, its contents should not be modified.
		
		Yields
		------
		ThingType
			The next Thing in the list of elements.
		"""
		for element in self.__ELEMENTS:
			yield element


	def __bool__(self):
		"""
		Indicating whether the View contains elements.
		
		Returns
		-------
		bool
			If the View is empty the resulting value is False and True otherwise.
		"""
		return len(self.__ELEMENTS) > 0


	@blue.restrict
	def __getitem__(self, key: int|slice|str) -> blue.ThingType|blue.ViewType:
		"""
		This method either returns a single Thing from the View or a new View from a subset of the View.
		
		Parameters
		----------
		key : int | slice | str
			If the argument is an integer, the Thing with its corresponding index is returned.
			If the argument is a slice, a View containing the Things according to the slices indicies is returned.
			if the argument is a string, a View containing the Things with the same user defined names is returned.
		
		Returns
		-------
		blue.ThingType | blue.ViewType
		
		Raises
		------
		NotImplemented
			If the argument is neither (int, slice, str) NotImplemented is raised.
		"""
		if isinstance(key, int):
			return self.__ELEMENTS[key]
		elif isinstance(key, slice):
			return View(self.__ELEMENTS[key], self.__NAME, self.__PARENT)
		elif isinstance(key, str):
			return View(list(filter(lambda x: x._name == key, self)), self.__NAME, self.__PARENT)
		else:
			raise NotImplemented


	def __getattr__(self, attr: str) -> list|blue.ViewType|blue.FunctionHandleType:
		"""
		This methods handles the access to the attributes of the Things contained in the View 
		depending on the type of attribute:
		1. *property* In this case a list of all the singularly retrieved attributes is returned.
		2. *children* This returns a new View, that contains the children of all the Views Things.
		3. *method/function* In this case a FunctionHandle is returned. 
		
		In the last case the :class:`FunctionHandle` can either be called directly and all arguments are 
		passed to each of the Views Thing or the distribute method of the FunctionHandle can be used, 
		to match each argument to each Thing in a one to one relation.
		
		.. code-block::
			:caption: Example
		
			>>> names   = ['thumb', 'index', 'middle', 'ring', 'pinky']
			>>> fingers = [blue.Body(name=name) for name in names]
			>>> palm = blue.Body(name='palm', bodies=fingers)
			>>> view = palm.bodies # calling child attributes creates a view
			>>> view
			View[5|palm.bodies]
			>>> view.name # CASE 1
			['thumb', 'index', 'middle', 'ring', 'pinky']
			>>> view.bodies # CASE 2
			View[0|palm.bodies.bodies]
			>>> intermediate = blue.Body(name='intermediate')
			>>> distal = blue.Body(name='distal')
			>>> phalanges = [distal, intermediate, intermediate, intermediate, intermediate]
			>>> view.attach.distribute(phalanges) # CASE 3
			>>> view.bodies[1:].attach(blue.Body(name='distal')) # CASE 3
		
		.. code-block:: mxml
			:caption: XML structure of the previous example
		
			<body name="palm">
			    <body name="thumb">
			        <body name="distal_(0)" />
			    </body>
			    <body name="index">
			        <body name="intermediate_(0)">
			            <body name="distal_(1)" />
			        </body>
			    </body>
			    <body name="middle">
			        <body name="intermediate_(1)">
			            <body name="distal_(2)" />
			        </body>
			    </body>
			    <body name="ring">
			        <body name="intermediate_(2)">
			            <body name="distal_(3)" />
			        </body>
			    </body>
			    <body name="pinky">
			        <body name="intermediate_(3)">
			            <body name="distal_(4)" />
			        </body>
			    </body>
			</body>
		
		
		Parameters
		----------
		attr : str
			All attributes are valid, aslong as they are properties or methods of the Thing type contained in the View.
		
		Returns
		-------
		list | blue.ViewType | blue.FunctionHandleType
		
		Raises
		------
		AttributeError
			If the attribute cannot be retrieved an Error is raised.
		"""
		if len(self) == 0:
			raise AttributeError(f'{self} is empty, attribute "{attr}" cannot be retrieved.')
		if all(hasattr(x, attr) for x in self):
			elements = list(map(lambda x: getattr(x, attr), self))
			if all(map(lambda x: isinstance(x, blue.ViewType), elements)):
				elements = list((item for view in elements for item in view))
				return View(elements, f'{self.__NAME}.{attr}', self.__PARENT)
			elif all(map(lambda x: isinstance(x, blue.ThingType), elements)):
				return View(elements, f'{self.__NAME}.{attr}', self.__PARENT)
			elif all(map(lambda x: isinstance(x, (FunctionType, MethodType)), elements)):
				return FunctionHandle(elements, f'{self.__NAME}.{attr}', self.__PARENT)
			elif all(map(lambda x: x is None, elements)):
				return [None] * len(elements) # []
			else:
				return elements
		else:
			try:
				#elements = list((item for view in map(lambda x: x.__getattr__(attr), self) for item in view))
				elements = list((item for view in map(lambda x: x.__getattribute__(attr), self) for item in view))
			except AttributeError as error:
				raise AttributeError(f'{self}.{attr} cannot be retrieved.') from None
			return View(elements, f'{self.__NAME}.{attr}', self.__PARENT)


	@blue.restrict
	def __setattr__(self, attr: str, val: object) -> None:
		"""
		This method passes the setting of attributes to each individual Thing contained in View.
		Depending on whether ``val`` is a list/View containing values for each Thing individually
		or a singular value, ``val`` is either unrolled along the first axis or given fully for each 
		assignment.

		>>> # keeping the view object from __getattr__
		>>> view.bodies[1:].name = ['intermediate_index', 
					    'intermediate_middle', 
					    'intermediate_ring', 
					    'intermediate_pinky']
		>>> view.bodies.name
		['intermediate_index', 
		 'intermediate_middle', 
		 'intermediate_ring', 
		 'intermediate_pinky']

		.. code-block:: mxml
			:caption: XML structure
		
			<body name="palm">
			    <body name="thumb">
			        <body name="distal_(0)" />
			    </body>
			    <body name="index">
			        <body name="intermediate_index">
			            <body name="distal_(1)" />
			        </body>
			    </body>
			    <body name="middle">
			        <body name="intermediate_middle">
			            <body name="distal_(2)" />
			        </body>
			    </body>
			    <body name="ring">
			        <body name="intermediate_ring">
			            <body name="distal_(3)" />
			        </body>
			    </body>
			    <body name="pinky">
			        <body name="intermediate_pinky">
			            <body name="distal_(4)" />
			        </body>
			    </body>
			</body>
		
		.. note::
			If values should be unrolled over all Things, ``vals`` must have the same length as the View!
		
		Parameters
		----------
		attr : str
			The name of the attribute to be set.
		val : object
			Either a list View or other iterable that contains different values for all children or one value of any type that is set for all.
		"""
		if attr in self.__RESERVED:
			return super().__setattr__(attr, val)
		else:
			view_attr = self.__getattr__(attr)
			for element, elem_val in zip(self, self.__nesting_value(view_attr, val)):
				element.__setattr__(attr, elem_val)


	def __delattr__(self, attr):
		"""
		This method detaches the attributes and deletes them afterwards
		"""
		view_attr = self.__getattr__(attr)
		if isinstance(view_attr, blue.ViewType):
			view_attr.parent.detach.distribute(view_attr)
			self.__PARENT._decouple_descendants()


	def __delitem__(self, key):
		"""
		This method detaches the attributes and deletes them afterwards
		"""
		view_key = self[key]
		if isinstance(view_key, blue.ViewType):
			view_key.parent.detach.distribute(view_key)
			view_key.parent._decouple_descendants()
			self.__PARENT._decouple_descendants()


	def apply(self, function, *args, **kwargs) -> None|list:
		"""
		Applies a function to all Things contained in View. Arguments and keyword arguments
		are passed to each individual call in an all to all relation.
		
		>>> # keeping the view object from __setattr__
		>>> renaming = lambda x: x.__setattr__('name', f'distal_{x.parent.parent.name}')
		>>> view.bodies.bodies.apply(renaming)
		>>> view.bodies.bodies.name
		['distal_index', 'distal_middle', 'distal_ring', 'distal_pinky']
		
		Parameters
		----------
		function : Callable
			Functions methods and classes are valid arguments.
		*args
			Arguments passed to each individual call.
		**kwargs
			Kexword arguments passed to each individual call.
		
		Returns
		-------
		None | list
			If any individual call results in a value that is not None, a list of all returned values gets returned, None otherwise.
		"""
		results = [function(element) for element in self]
		if any(x is not None for x in results):
			return results


	def __nesting_value(self, view_attr, val):
		"""
		Thie method is used to create an iterator for value assignment that either creates 
		a many to one or a one to one relation, depending on the new values shape and the 
		current attributes shape.
		"""
		view_structure = self.__nesting_structure(view_attr)
		val_structure  = self.__nesting_structure(val)
		if isinstance(val, np.ndarray) and  self.__compatible_numpy(view_structure, val):
			return np.rollaxis(val, axis=0)
		elif self.__compatible_nesting(view_structure, val_structure):
			return val
		elif isinstance(val, blue.ViewType):
			if len(val) != 1:
				raise ValueError(f'To set a View attribute to another View both sizes must be identical or the other Views size must be one, got sizes {len(view_structure)} and {len(val_structure)}.')
			else:
				return (val[0] for _ in range(len(self)))
		else:
			return (val for _ in range(len(view_structure)))



	@classmethod
	def __nesting_structure(cls, element):
		"""
		This method retrieves a representation of the elements nesting structure.
		"""
		if isinstance(element, np.ndarray):
			return element.shape
		elif isinstance(element, str):
			return str
		elif isinstance(element, View):
			return cls.__nesting_structure(list(element))
		elif hasattr(element, '__iter__'):
			return element.__class__(map(cls.__nesting_structure, element))
		else:
			return element.__class__


	@staticmethod
	def __compatible_nesting(view_structure, val_structure):
		"""
		This method determines whether the two nesting structures are compatible for a 
		one to one assignment relation for lists and other non numpy iterables.
		"""
		val_class = val_structure if isinstance(val_structure, type) else val_structure.__class__
		if not hasattr(val_structure, '__len__') or issubclass(val_class, str):
			return False
		elif len(val_structure) != len(view_structure):
			return False
		elif (not hasattr(val_structure, '__contains__') or view_structure not in val_structure) and val_structure != view_structure:
			return False
		else:
			return True


	@staticmethod
	def __compatible_numpy(view_structure, val):
		"""
		This method determines whether the two nesting structures are compatible for a 
		one to one assignment relation for numpy arrays.
		"""
		if val.shape[0] == len(view_structure) and val.ndim > 1:
			if all(isinstance(view, tuple) and view == val.shape[1:] for view in view_structure):
				return True
			elif all(isinstance(view, (list, tuple)) and len(view) == val.shape[1] for view in view_structure):
				return True
		return False



class FunctionHandle(blue.FunctionHandleType):

	"""
	This class is used to handle access to a function or method called from a :class:`View`. 

	>>> palm.bodies
	View[5|palm.bodies]
	>>> palm.bodies.shift
	ViewFunction[5|palm.bodies.shift]
	
	Arguments given to the function or method call are passed to each individual Thing
	in a many to one relation.

	>>> palm.bodies.pos
	[array([0., 0., 0.], dtype=float32), 
	 array([0., 0., 0.], dtype=float32), 
	 array([0., 0., 0.], dtype=float32), 
	 array([0., 0., 0.], dtype=float32), 
	 array([0., 0., 0.], dtype=float32)]
	>>> many_to_one = palm.bodies.shift([2, 0, -1])
	>>> many_to_one
	View[5|palm.bodies.shift(...)]
	>>> many_to_one.pos
	[array([ 2.,  0., -1.], dtype=float32), 
	 array([ 2.,  0., -1.], dtype=float32), 
	 array([ 2.,  0., -1.], dtype=float32), 
	 array([ 2.,  0., -1.], dtype=float32), 
	 array([ 2.,  0., -1.], dtype=float32)]

	If the :meth:`distributed` over all Things contained in the parenting
	View in a one to one relation by using the distribute submethod.

	>>> one_to_one = palm.bodies.shift.distribute([[0, 0, 1], 
						       [0, 1, 0], 
						       [0, 1, 1], 
						       [1, 0, 0], 
						       [1, 0, 1]])
	>>> one_to_one.pos
	[array([0., 0., 1.], dtype=float32), 
	 array([0., 1., 0.], dtype=float32), 
	 array([0., 1., 1.], dtype=float32), 
	 array([1., 0., 0.], dtype=float32), 
	 array([1., 0., 1.], dtype=float32)]

	If the result of all method or function calls are None, None is returned.

	>>> palm.bodies.attach()
	None

	If all results are Things, a View of them is returned. This allows for 
	functions and methods that return Things to be chained.

	>>> palm.bodies.copy()
	View[5|palm.bodies.copy(...)]
	>>> palm.bodies.shift([0, 0, 3]).rotate(alpha=PI).copy()
	View[5|palm.bodies.shift(...).rotate(...).copy(...)]

	Otherwise a list of all individually returned values is returned. 

	>>> palm.bodies._mujoco_specs()
	[{'name': 'thumb'}, 
	 {'name': 'index'}, 
	 {'name': 'middle'}, 
	 {'name': 'ring'}, 
	 {'name': 'pinky'}]

	"""
	
	def __init__(self, functions: list, name: str, parent: blue.ThingType) -> None:
		"""
		The FunctionHandle takes the functions or methods to be called, the name by which 
		it was called and the initial parent from which the first :class:`View` was called.
		
		Parameters
		----------
		functions : list
			All functions or methods from the Views Things.
		name : str
			The name of the called function or method.
		parent : blue.ThingType
			The parent from which the first View was called.
		"""
		self.__NAME      = name
		self.__PARENT    = parent
		self.__FUNCTIONS = functions


	def __repr__(self) -> str:
		"""
		Represents the View as a string.
		
		Returns
		-------
		str
			The representing string contains a representation of the call that was used to get the View.
		"""
		return f'ViewFunction[{len(self)}|{self.__PARENT.name}.{self.__NAME}]'


	def __iter__(self):
		"""
		Yields
		------
		Callable
			The functions or methods Things that were requested from the View.
		"""
		for functions in self.__FUNCTIONS:
			yield functions


	def __len__(self) -> int:
		"""
		Returns
		-------
		int
			The number of 
		"""
		return len(self.__FUNCTIONS)


	def __call__(self, *args, **kwargs) -> None|list:
		"""
		The call is redirected to each function or method in the FunctionHandle.
		All arguments are passed on all calls in a many to one relation.
		
		Parameters
		----------
		*args
			The arguments to be passed to the call.
		**kwargs
			The keywordarguments to be passed to the call.
		
		Returns
		-------
		None | list | View
			The aggregated result of each call.
		"""
		result = [functions(*args, **kwargs) for functions in self]
		return self._convert_result(result)


	def distribute(self, *args, **kwargs) -> None|list:
		"""
		The call is redirected to each function or method in the FunctionHandle.
		The arguments are unrolled along the first axis and distributed over all calls 
		in a one to one relation.
		
		Parameters
		----------
		*args
			The arguments to be passed to the call.
		**kwargs
			The keywordarguments to be passed to the call.
		
		Returns
		-------
		None | list | View
			The aggregated result of each call.
		"""
		kw_list = [dict(zip(kwargs.keys(), vals)) for vals in zip(*kwargs.values())]
		if args and kwargs:
			assert all(len(self) == l for l in map(len, args))
			assert len(self) == len(kw_list)
			result = [function(*arg, **kwarg) for function, arg, kwarg in zip(self, zip(*args), kw_list)]
		elif args:
			assert all(len(self) == l for l in map(len, args))
			result = [function(*arg) for function, arg in zip(self, zip(*args))]
		elif kwargs:
			assert len(self) == len(kw_list)
			result = [function(**kwarg) for function, kwarg in zip(self, kw_list)]
		else:
			result = [function() for function in self]
		return self._convert_result(result)


	def _convert_result(self, result):
		if any(x is not None for x in result):
			if all(isinstance(x, blue.ThingType) for x in result):
				return View(result, f'{self.__NAME}(...)', self.__PARENT)
			else:
				return result
		else:
			return None



class AllView(blue.AllViewType, View):
	"""
	The AllView is analogous to the :class:`View` for all descendants instead of children.
	Attribute getting and setting is handeled for all valid descendants. A ThingsType 
	attribute is then converted back to a normal view.

	.. code-block:: mxml
		:caption: XML Structure:

		<body name="A">
                        <body name="B" />
                        <geom mass="42.0" type="sphere" name="C" />
                </body>

	.. code-block:: python
		:caption: Children and Descendants View Comparison:

		>>> world.bodies
		View[1|anonymous_model.bodies]
		>>> world.bodies.name
		['A']
		>>> world.all
		View[3|anonymous_model.all]
		>>> world.all.name
		['A', 'B', 'C']
		>>> world.all.bodies
		View[2|anonymous_model.all.bodies]
		>>> world.bodies.name
		['A', 'B']

	If only some descendants have a certain attribute, the view restricts to this valid subset. 
	For example, only :class:`Geoms <blueprints.geoms.BaseGeom>` have mass, so retrieving ``all.mass`` excludes all non-Geoms.

	.. code-block:: python
		:caption: Children and Descendants View Comparison:

		>>> world.all.mass
		[42.0]
	"""
	__RESERVED = {'__ELEMENTS', 
		      '__NAME', 
		      '__PARENT', 
		      '__RESERVED', 
		      '__DESCENDANTS', 
		      '_View__ELEMENTS', 
		      '_View__NAME', 
		      '_View__PARENT', 
		      '_View__RESERVED', 
		      '_AllView__ELEMENTS', 
		      '_AllView__NAME', 
		      '_AllView__PARENT', 
		      '_AllView__RESERVED', 
		      '_AllView__DESCENDANTS'}
	@blue.restrict
	def __init__(self, descendants: dict, parent: blue.ThingType) -> None:
		"""
		Parameters
		----------
		descendants : dict
			The descendants property invoked on ``parent``
		parent : blue.ThingType
			The parent Thing from which ``.all`` was retrieved.
		"""
		elements = []
		for desc_dict in descendants.values():
			elements.extend(desc_dict['descendants'])
		super().__init__(elements=elements, name='all', parent=parent)
		self.__DESCENDANTS = descendants


	def __getattr__(self, attr: str):
		if attr in self.__RESERVED:
			return super().__getattribute__(attr)
		if attr in self.__DESCENDANTS:
			return View(self.__DESCENDANTS[attr]['descendants'], f'{self._View__NAME}.{attr}', self._View__PARENT)
		else:
			if not all(hasattr(thing, attr) for thing in self):
				elements = list(filter(lambda x: hasattr(x, attr), self))
				view = View(elements, name=self._View__NAME, parent=self._View__PARENT)
				return view.__getattr__(attr)
			else:
				return super().__getattr__(attr)


	@blue.restrict
	def __setattr__(self, attr: str, val: object) -> None:
		super().__setattr__(attr, val)
		if attr in self.__RESERVED or all(hasattr(thing, attr) for thing in self):
			super().__setattr__(attr, val)
		else:
			elements = list(filter(lambda x: hasattr(x, attr), self))
			view = View(elements, name=self._View__NAME, parent=self._View__PARENT)
			view.__setattr__(attr, val)



class LatticeView(blue.LatticeViewType):
	"""
	A LatticeView is created whenever a :class:`Lattice <blueprints.utils.lattice.Lattice>` is indexed. 
	Attribute getter and setter are handled through this View similar to :class:`View`. For details have 
	a look at :class:`Lattice <blueprints.utils.lattice.Lattice>`.
	"""
	__RESERVED = {'_LatticeView__THINGS', 
		      '_LatticeView__KEYS', 
		      '_LatticeView__KEY_STR', 
		      '_LatticeView__PARENT'}
	def __init__(self, 
		     things: list, 
		     parent: blue.LatticeType, 
		     keys:   tuple[int|slice] = ()):
		"""
		Parameters
		----------
		things : list
			The indexed Things of the parent Lattice
		parent : blue.LatticeType
			The Lattice which was indexed to create the LatticeView
		keys : tuple[int | slice]
			The keys used to index the Lattice
		"""
		def parse(key):
			if isinstance(key, int):
				return str(key)
			elif isinstance(key, slice):
				indecies = []
				if key.step is not None:
					indecies.insert(0, str(key.step))
				if key.stop is not None:
					indecies.insert(0, str(key.stop))
				else:
					indecies.insert(0, '')
				if key.start is not None:
					indecies.insert(0, str(key.start))
				else:
					indecies.insert(0, '')
				return ':'.join(indecies)
			else:
				raise NotImplemented
		self.__KEY_STR = ', '.join(map(parse, keys))
		self.__KEYS    = keys
		self.__THINGS  = things if isinstance(things, list) else [things]
		self.__PARENT  = parent


	@property
	def n_dim(self) -> int:
		"""
		The number of dimensions in the Lattice
		
		Returns
		-------
		int
		"""
		things = self.__THINGS
		n = 0
		while isinstance(things, list):
			n += 1
			things = things[0]
		return n


	@blue.restrict
	def __getitem__(self, keys: int|slice|tuple[int|slice]) -> blue.LatticeViewType|blue.MoveableThingType:
		if isinstance(keys, tuple) and len(keys) > self.n_dim:
			raise IndexError(f'Lattice {self} has {self.n_dim} axes but {len(keys)} were indexed.')
		if isinstance(keys, (int, slice)):
			keys = (keys,)
		def rec(things, keys):
			if len(keys) == 0:
				return things
			else:
				key, keys = keys[0], keys[1:]
				if isinstance(key, int):
					return rec(things[key], keys)
				else:
					return [rec(thing, keys) for thing in things[key]]
		if all(isinstance(key, int) for key in keys) and len(keys) == self.n_dim:
			return rec(self.__THINGS, keys)
		return blue.LatticeView(rec(self.__THINGS, keys), self.__PARENT, keys)


	def __iter__(self) -> blue.ThingType:
		queue = deque(self.__THINGS)
		while queue:
			item = queue.popleft()
			if isinstance(item, list):
				queue.extendleft(reversed(item))
			else:
				yield item


	def __len__(self) -> int:
		"""
		Returns
		-------
		int
			The number of Things contained in the LatticeView.
		"""
		things = self.__THINGS
		l = 1 if things else 0
		while isinstance(things, list):
			l *= len(things)
			things = things[0]
		return l


	def __repr__(self) -> str:
		"""
		Represents the View as a string.
		
		Returns
		-------
		str
			The representing string contains a representation of the call that was used to get the View.
		"""
		if self.__KEY_STR:
			return f'LatticeView[{self.__KEY_STR}]'
		else:
			return 'LatticeView'


	def __getattr__(self, attr):
		if len(self) == 0:
			raise AttributeError(f'{self} is empty, attribute "{attr}" cannot be retrieved.')
		if all(hasattr(x, attr) for x in self):
			things = self._rec(lambda x: getattr(x, attr), list, self.__THINGS)
			if self._rec(lambda x: isinstance(x, blue.ViewType), all, things):
				things = list(self._rec(list, lambda x: list(chain(*x)), things))
				return View(things, attr, self.__PARENT)
			elif self._rec(lambda x: isinstance(x, blue.ThingType), all, things):
				things = list(self._rec(lambda x: [x], lambda x: list(chain(*x)), things))
				return View(things, attr, self.__PARENT)
			elif self._rec(lambda x: isinstance(x, (FunctionType, MethodType)), all, things):
				things = list(self._rec(lambda x: [x], lambda x: list(chain(*x)), things))
				return FunctionHandle(things, attr, self.__PARENT)
			elif self._rec(lambda x: x is None, all, things):
				return [None] * len(self) # []
			else:
				return things
		else:
			try:
				things = self._rec(lambda x: x.__getattribute__(attr), list, self.__THINGS)
			except AttributeError as error:
				raise AttributeError(f'{self}.{attr} cannot be retrieved.') from None
			return View(things, attr, self.__PARENT)


	def __setattr__(self, attr, value):
		if attr in self.__RESERVED:
			super().__setattr__(attr, value)
		else:
			view_attr = self.__getattr__(attr)
			view_structure = self.__nesting_structure(view_attr)
			val_structure  = self.__nesting_structure(value)
			if view_structure == val_structure:
				def rec(func, things, values):
					if isinstance(things, list):
						for thing, value in zip(things, values):
							rec(func, thing, value)
					else:
						func(things, values)
				rec(lambda thing, value: thing.__setattr__(attr, value), self.__THINGS, value)
			else:
				def rec(func, things):
					if isinstance(things, list):
						for thing in things:
							rec(func, thing)
					else:
						func(things)
				rec(lambda thing: thing.__setattr__(attr, value), self.__THINGS)


	@classmethod
	def _rec(cls, func, aggregate, things):
		"""
		This classmethod is a helper to apply a function on individual Things and aggregate the results. 
		"""
		if isinstance(things, list):
			return aggregate(cls._rec(func, aggregate, thing) for thing in things)
		else:
			return func(things)


	@classmethod
	def __nesting_structure(cls, element):
		"""
		This method retrieves a representation of the elements nesting structure.
		"""
		if isinstance(element, np.ndarray):
			return element.shape
		elif isinstance(element, str):
			return str
		elif isinstance(element, View):
			return cls.__nesting_structure(list(element))
		elif hasattr(element, '__iter__'):
			return element.__class__(map(cls.__nesting_structure, element))
		else:
			return element.__class__



import numpy as np
from math import trunc, ceil
from itertools import product



def smoothstep(x):
	y = 3 * x**2 - 2 * x**3
	y[x <= 0] = 0.
	y[x >= 1] = 1.
	return y



def padding(x, axis):
	axis = tuple(slice(None if i == -1 else 1, -1 if i == -1 else None) for i in axis)
	return x[*axis,...]



def resize(img, shape):
	x_shape = img.shape
	y_shape = shape
	for x, y in zip(x_shape, y_shape):
		X = np.arange(x)[None,...]
		Y = np.arange(y)[...,None]
		mini = np.minimum((1 + X) * y / x, (1 + Y))
		maxi = np.maximum(     X  * y / x,      Y )
		M = np.maximum(0, mini - maxi)
		img = np.einsum('ij,j...->i...', M, img)
		img = np.moveaxis(img, 0, -1)
	return img



def perlin(resolution, frequency, periodic=False):
	ALPHA = 'abcdefghijklmnopqrstuvwx'
	min_width = int(ceil(min(resolution) / frequency))
	grid_size = tuple(int(ceil(x / min_width)) for x in resolution)
	cell_size = tuple(int(ceil(x / y)) for x, y in zip(resolution, grid_size))
	ndim      = len(grid_size)
	cell      = np.stack(np.meshgrid(*(np.linspace(-1, 1, n) for n in cell_size)), axis=-1)
	corners   = list(map(np.array, product(*((-1, 1) for _ in range(ndim)))))
	offsets   = np.stack(list(map(lambda x: cell - x, corners)), axis=-1)
	distances = np.sqrt(np.sum(offsets**2, axis=-2)) / 2
	factors   = 1 - smoothstep(distances)
	grads     = np.random.uniform(low=-1, high=1, size=(*(x + 1 for x in grid_size), ndim))
	grads     = grads / (1e-8 + np.sqrt(np.sum(grads**2, axis=-1)[...,None]))
	if periodic:
		for i, _ in enumerate(grid_size):
			grads[*(slice(None) for _ in range(i)), -1, ...] = grads[*(slice(None) for _ in range(i)),0,...]
	rule     = f'{ALPHA[:ndim].upper()}y,{ALPHA[:ndim]}yz->{ALPHA[:ndim].upper()}{ALPHA[:ndim]}z'
	products = np.einsum(rule, grads, offsets)
	correlations = np.stack([padding(products[...,i], corner) for i, corner in enumerate(corners)], axis=-1)
	rule     = f'{ALPHA[:ndim].upper()}{ALPHA[:ndim]}y,{ALPHA[:ndim]}y->{ALPHA[:ndim].upper()}{ALPHA[:ndim]}'
	heights  = np.einsum(rule, correlations, factors)
	heights  = np.moveaxis(heights, 0, 1)
	for _ in range(ndim):
		heights = np.concatenate(np.rollaxis(heights, axis=0), axis=ndim-1)
	heights  = np.moveaxis(heights, 0, 1)
	heights  = resize(heights, resolution)
	return heights



import blueprints as blue
from blueprints import restrict
from . import base

import numpy as np
import inspect
import xml.etree.ElementTree as xml
from copy import copy
from collections import defaultdict
from itertools import chain



class NodeThing(base.BaseThing, blue.NodeThingType):
	"""
	If an inheriting Thing class can have children attached to it in the kinematic 
	tree it derives from NodeThing. It contains the methods :meth:`NodeThing.attach` and 
	:meth:`NodeThing.detach` can be used add or detach children. 
	
	If a Thing class inherits from :class:`NodeThing` it needs to define a dictionary
	named ``_CHILDREN`` which has the names of children attributes as keys and as values 
	a second level dictionary containing the ThingType of the respective children under 
	the keys ``'type'`` and a reference to the list in which the children are stored under 
	the key ``'children'``.
	"""
	@restrict
	def _build(self, parent, world, indicies, **kwargs):
		"""
		This method is called by the Things parent to construct the xml representations of 
		the kinematic tree. It recursively calls the :meth:`BaseThing._build` methods if its 
		children.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The built xml element of the Thing.
		"""
		self._xml_root = xml.SubElement(parent, 
						self._MUJOCO_OBJ, 
						**self._mujoco_specs(kwargs))
		# BUILDING CHILDREN
		self._build_children(parent=self._xml_root, 
				     world=world, 
				     indicies=indicies)
		return self._xml_root

	# KINEMATIC TREE METHODS/PROPERTIES

	def __getattr__(self, attr: str) -> blue.View|object|None:
		"""
		Attributes that handle access to the children of a :class:`NodeThing` are 
		defined implicitly in :attr:`NodeThing._CHILDREN`. If the attribute name can 
		be retrieved from :attr:`NodeThing._CHILDREN` a :class:`blueprints.utils.view.View` is 
		returned.
		
		If the attribute cannot be obtained from :attr:`NodeThing._CHILDREN` the call is 
		passed to ``super().__getattr__`` if is it exists.
		
		Parameters
		----------
		attr : str
			Name of the attribute that is requested.
		
		Returns
		-------
		blue.View | object | None
			Either a View to the requested children is returned, or a value obtained from 
			``super().__getattr__``.
		
		Raises
		------
		AttributeError
			If the attribute does not exist an Error is raised.
		"""

		# THIS IS MAGIC! DON'T TOUCH
		# Reserved attributes _CHILDREN and _PSEUDO_CHILDREN might not exist always (for example
		# during init). The cunjunction with RESERVED in the boolean flag compute ensures, that 
		# no infinite recursion is triggered if getattr on _CHILDREN or _PSEUDO_CHILDREN is called.
		# BEGIN MAGIC
		RESERVED = attr not in {'_CHILDREN', '_PSEUDO_CHILDREN'}#, 'descendants', 'children'}
		CHILDREN = RESERVED and attr in self._CHILDREN
		PSEUDO_C = RESERVED and hasattr(self, '_PSEUDO_CHILDREN') and attr in self._PSEUDO_CHILDREN
		# END MAGIC
		if CHILDREN or PSEUDO_C:
			if attr in self._CHILDREN:
				 elements = self._CHILDREN[attr]['children']
			else:
				 elements = self._PSEUDO_CHILDREN[attr]['children']
			return blue.View(elements=elements, 
					 name=attr, 
					 parent=self)
		#if attr != '_CHILDREN' and attr in self._CHILDREN:
		#	elements = self._CHILDREN[attr]['children']
		#	return blue.View(elements=elements, 
		#			 name=attr, 
		#			 parent=self)
		elif hasattr(super(), '__getattr__'):
			return super().__getattr__(attr)
		else:
			raise AttributeError(f'{type(self)} does not have attribute {attr}')


	def __setattr__(self, attr: str, val: object) -> None:
		"""
		If a valid children name is set, val must be either :class:`list` or 
		:class:`blueprints.utils.view.View`. The children contained in val are 
		then attached to the Thing and the previous children attached under the 
		attribute are detached.
		
		Parameters
		----------
		attr : str
			Name of the attribute that is set.
		val : object
			The value that is to be assigned to the attribute.
		
		Raises
		------
		ValueError
			If val is not :class:`list` or :class:`blueprints.utils.view.View` an Error is raised.
		"""
		if hasattr(self, '_CHILDREN') and attr in self._CHILDREN:
			if not isinstance(val, (list, blue.ThingType, blue.View)):
				raise ValueError(f'Setting the children of a object is only possible for Thing, list, or View, got {type(val)}.')
			children = self.__getattr__(attr)
			val      = [val] if isinstance(val, blue.ThingType) else val
			self.detach(*children)
			self.attach(*val, copy=False)
		else:
			super().__setattr__(attr, val)


	@restrict
	def copy(self, 
		 shallow: bool = False, 
		 **kwargs) -> blue.NodeThingType:
		"""
		This method creates a copy of the Thing from which it was called. Additional 
		modifications to the copy can be set as keyword arguments. If shallow is True 
		only the Thing itself, otherwise its children are also copied and attached.
		
		Parameters
		----------
		shallow : bool, optional
			Indicating whether the children should be copied and attached as well.
		**kwargs
			Specifies additional attributes that are passed to the copied 
			:meth:`NodeThing.__init__`.
		
		Returns
		-------
		blue.NodeThingType
			The returned value is a copy of the :class:`ǸodeThing` itself optionally 
			including copies or its children.
		"""
		if blue.REGISTER.copy_root is None:
			blue.REGISTER.copy_root = self
		if shallow:
			children        = {name: None for name in self._CHILDREN.keys()}
			single_children = {name: None for name in self._SINGLE_CHILD_ATTR}
		elif hasattr(self, '_CHILDREN'):
			children  = {name: [thing.copy() for thing in list(child_dict['children'])] \
							 for name, child_dict in self._CHILDREN.items() \
							 if name not in kwargs and not issubclass(child_dict['type'], blue.CyclicalThingType)}
			cyclicals = {name: [thing        for thing in list(child_dict['children'])] \
							 for name, child_dict in self._CHILDREN.items() \
							 if name not in kwargs and     issubclass(child_dict['type'], blue.CyclicalThingType)}
			single_children = {name: self.__getattribute__(name).copy() if name not in self._NO_COPY_ATTR() else self.__getattribute__(name) \
										    for name in self._SINGLE_CHILD_ATTR() \
										    if self.__getattribute__(name) is not None}
			pack = lambda x: x if isinstance(x, list) else [x]
			copy = lambda x: x.copy()
			kwargs_children = {name: list(map(copy, pack(val))) for name, val in kwargs.items() if name in self._CHILDREN or name in self._SINGLE_CHILD_ATTR()}
			kwargs.update(kwargs_children)
		else:
			children        = {}
			cyclicals       = {}
			single_children = {}
		blueprint_specs = self._blueprint_specs()
		blueprint_specs.update(children)
		blueprint_specs.update(single_children)
		blueprint_specs.update(kwargs)
		if 'name' not in kwargs:
			blueprint_specs['name'] = self._name
		thing = self.__class__(copy=False, **blueprint_specs)
		if not shallow and hasattr(self, '_CHILDREN'):
			self._migrate_children(thing, cyclicals)
			#self._migrate_children(thing, children, cyclicals)
		if blue.REGISTER.copy_root is self:
			blue.REGISTER.copy_root = None
		return thing


	@restrict
	def _migrate_children(self, thing, cyclicals):
		"""
		This methods finalizes the migration for cyclical kinematic hierachies.
		
		Parameters
		----------
		thing : NodeTingType
			The Thing to which self is migrating.
		#children : dict
		#	A dictionary with all acyclical children ThingTypes in the format of 
		#	:attr:`children`.
		cyclicals : dict
			A dictionary with all cyclical children ThingTypes in the format of 
			:attr:`children`.
		"""
		for children in cyclicals.values():
			for child in children:
				child._migrate('parent', thing)
		for child_dict in self._CYCLE_REF.values():
			for cyclical in list(child_dict['children']):
				cyclical._migrate(cyclical._PARENT_REFERENCE, thing)
		for child_dict in self._CHILDREN.values():
			if issubclass(child_dict['type'], blue.FocalThingType):
				for focal in child_dict['children']:
					focal._migrate(self, thing)


	@restrict
	def attach(self, 
		   *items:   list[blue.ThingType|blue.LatticeType|blue.ViewType], 
		   globally: bool = False, 
		   copy:     bool = True) -> None:
		"""
		This method attaches ``*items`` to the parent Thing. If copy is set true, a copy of the items is created. 
		If copy is set to false, the kinematic graph might no longer be a tree resulting in infinite loops 
		or naming conflicts if the same item is included in the tree twice. All items of a specific type 
		can be accessed via the types attribute.
		
		Parameters
		----------
		*items : list[blue.ThingType]
			All items that are to be attached.
		globally : bool
			A flag specifing whether the Things are attached in the local or global position and orientation.
		copy : bool, optional
			A flag indicating whether a copy of the items should be attached. It is recommended for the user to not set this argument to false.
		
		Raises
		------
		TypeError
			If an item is not a valid child an error is raised.
		"""
		#if self._launched:
		#	raise Exception('No Things can be attached after the World has been built. Attach all Things before build or use World.unbuild().')
		views   =                   list(chain( *filter(lambda x:     isinstance(x, blue.ViewType),                     items)))
		lattice =                   list(chain( *filter(lambda x:     isinstance(x, blue.LatticeType),                  items)))
		items   = views + lattice + list(        filter(lambda x: not isinstance(x, (blue.ViewType, blue.LatticeType)), items))
		for item in items:
			if copy:
				child = item.copy()
			else:
				child = item
				if isinstance(child, blue.NodeThing):
					if child in self.path:
						raise ValueError(f'Circular Attachment. {repr(child)} is a parent of {repr(self)}. Consider setting copy=True.')
			if child.parent is not None:
				child.parent.detach(child)
			# GLOBALLY
			transform = globally and isinstance(child, blue.MoveableThingType) and isinstance(self, blue.MoveableThingType)
			if transform:
				R1  = child.global_rotation_matrix
				R2  = self.global_rotation_matrix
				R3  = R2.T @ R1
				euler = blue.Rotation.reference_frame_to_euler(R3)
				pos = child.global_pos
			types = tuple(map(lambda x: x['type'], filter(lambda x: not issubclass(x['type'], blue.TendonType), self._CHILDREN.values())))
			if not isinstance(child, types):
				raise TypeError(f'{type(self)} only takes {types} for attachment got {type(child)} instead.')
			for name, child_dict in self._CHILDREN.items():
				child_type = child_dict['type']
				children   = child_dict['children']
				if isinstance(child, child_type):
					children.append(child)
					child._parent = self
					#child.parent = self
					break
			#child._name_manager = None
			if transform:
				child.global_pos = pos
				child.euler = list(euler)


	@restrict
	def detach(self, *items: list[blue.ThingType]) -> None:
		"""
		This method is used to detach Things from the kinematic tree. They are afterwards 
		no longer present in this Things children attributes and their ``parent`` attribute 
		will be set to ``None``. 
		
		Parameters
		----------
		*items
			Things that are no longer children of this Thing can be passed, as long as their 
			type is a valid type for children of this :class:`NodeThing`.
		
		Raises
		------
		TypeError
			If the arguments to this function are not valid children types an error is raised.
		"""
		if self._launched:
			raise Exception('No Things can be detached after the World has been built. Detach all Things before build or use World.unbuild().')
		views = list(chain( *filter(lambda x:     isinstance(x, blue.ViewType), items)))
		items = views + list(filter(lambda x: not isinstance(x, blue.ViewType), items))
		for item in items:
			#types = tuple(map(lambda x: x['type'], self._CHILDREN.values()))
			#if not isinstance(item, types):
			#	raise TypeError(f'{type(self)} only takes {types} or Views of those types for detachment got {type(item)} instead.')
			# FIND THE CHILD IN CHILDREN
			for name, child_dict in self._CHILDREN.items():
				child_type = child_dict['type']
				children   = child_dict['children']
				if isinstance(item, child_type):
					if item in children:
						# REMOVE CHILD
						children.remove(item)
						#item._name_scope = None
						item._parent = None
						#item.parent = None
						# DECOUPLE ORPHANED REFERENCES OF CYCLICAL THINGS
						if isinstance(item, blue.NodeThingType):
							#item._name_manager = blue.naming.NameManager(item)
							item._decouple_descendants()
						if isinstance(item, blue.CyclicalThingType):
							item._decouple()
			# FIND REF IN CYCLE REFS
			for name, child_dict in self._CYCLE_REF.items():
				child_type = child_dict['type']
				children   = child_dict['children']
				if isinstance(item, child_type):
					if item in children:
						# REMOVE CHILD
						children.remove(item)
						#item._name_scope = None
						item.__setattr__(item._PARENT_REFERENCE, None)
		self._decouple_descendants()


	def __contains__(self, thing: blue.ThingType) -> bool:
		"""
		Parameters
		----------
		thing : ThingType
			On checking whether the thing is a child instance identity is used rather 
			than a weaker attribute equivalence.
		
		Returns
		-------
		bool
			Indicates if the thing is currently a child of this :class:`NodeThing`.
		"""
		for child_dict in self._CHILDREN.values():
			child_type = child_dict['type']
			children   = child_dict['children']
			if isinstance(thing, child_type):
				if thing in children:
					return True
		return False


	@property
	def _location_range(self) -> tuple[np.ndarray, np.ndarray]|None:
		"""
		This function is used to determine the range in which the Thing and its 
		children are located.
		
		Returns
		-------
		tuple[np.ndarray, np.ndarray]
			The returned values are two numpy arrays of shape (3,) containing the 
			minimum and maximum of the coordinate values of each dimension which 
			contain all points of the Thing including its children.
		"""
		min_values, max_values = [], []
		if isinstance(self, blue.MoveableThingType):
			min_pos, max_pos = blue.MoveableThing._location_range.fget(self)
			min_values.append(min_pos)
			max_values.append(max_pos)
		for dec_dict in self.descendants.values():
			for decendant in dec_dict['descendants']:
				if hasattr(self, '_background'):
					if decendant in self._background:
						continue
				if hasattr(decendant, 'pos') and hasattr(decendant, 'size'):
					min_pos, max_pos = decendant._location_range
					min_values.append(min_pos)
					max_values.append(max_pos)
		if min_values and max_values:
			min_pos = np.min(np.array(min_values), axis=0)
			max_pos = np.max(np.array(max_values), axis=0)
			return min_pos, max_pos
		else:
			return np.zeros(3), np.zeros(3)


	@property
	def children(self) -> dict:
		"""
		The names and lists of all types of children are specified in the :attr:`_CHILDREN` attribute 
		as well as the type. This property is a reduced dictionary only containing names and lists.
		
		Returns
		-------
		dict
			A dictionary containing all children of the NodeThing. 
		"""
		return {name: child['children'] for name, child in self._CHILDREN.items()}


	@property
	def descendants(self) -> dict:
		"""
		Nested dictionaries — The top level contains the names of the respective 
		descendants property names, on the second level there are the keys ``'type'`` 
		and ``'descendant'`` containing the ThingType and the descendant of the attribute.
		
		Returns
		-------
		dict
			The structure of descendants differs from :attr:`children`.
		"""
		descendants = defaultdict(lambda: {'type':        None, 
						   'descendants': list()})
		# SINGLE CHILDREN
		queue = list((self, key, val) for key, val in self._SINGLE_CHILD_ATTR().items())
		while queue:
			thing, name, child_type = queue.pop()
			single = thing.__getattribute__(name)
			if single is None:
				continue
			descendants[name]['type'] = child_type
			descendants[name]['descendants'].append(single)
			queue += list((single, key, val) for key, val in single._SINGLE_CHILD_ATTR().items())
		# CHILDREN
		for name, child_dict in self._CHILDREN.items():
			child_type = child_dict['type']
			children   = child_dict['children']
			descendants[name]['type'] = child_type
			descendants[name]['descendants'].extend(children)
			if issubclass(child_type, blue.NodeThingType):
				for child in children:
					for dec_name, dec_dict in child.descendants.items():
						dec_type     = dec_dict['type']
						dec_children = dec_dict['descendants']
						descendants[dec_name]['type'] = dec_type
						#for dec_child in dec_children:
						#	if dec_child not in descendants[dec_name]['descendants']:
						#		descendants[dec_name]['descendants'].append(dec_child)
						descendants[dec_name]['descendants'].extend(dec_children)
		# ELIMINATE DUPLICATES (order-preserving, O(n))
		for dec_name, dec_dict in descendants.items():
			dec_dict['descendants'] = list(dict.fromkeys(dec_dict['descendants']))
		return descendants


	@property
	def all(self) -> blue.AllViewType:
		"""
		This property gives access to a :class:`View <blueprints.utils.view.AllView>` 
		through which attributes of all descendants can be handled. See :class:`AllView <blueprints.utils.view.AllView>` for details.

		Returns
		-------
		AllViewType
			A handle for accessing attributes of all descendants of the NodeThing.
		"""
		return blue.AllView(self.descendants, self)


	def _check_children_types(self):
		"""
		This method should be called at the end of a Thing class deriving from :class:`NodeThing` 
		:meth:`__init__`. It checks is all children have the proper type.
		
		Raises
		------
		TypeError
			If a child does not have the correct type and error is raised.
		"""
		for child_dict in self._CHILDREN.values():
			child_type = child_dict['type']
			children   = child_dict['children']
			if issubclass(child_type, blue.BodyType):
				for child in children:
					if type(child) == blue.Agent:
						raise TypeError(f"""A type restriction in function {self.__init__.__name__} was violated: 
An argument for bodies is was of type Agent. 
The received argument for {child.__class__.__name__.lower()} was {children}.""")
			for child in children:
				if not isinstance(child, child_type):
					raise TypeError(f"""A type restriction in function {self.__init__.__name__} was violated: 
{child.__class__.__name__.lower()} is supposed to be of type {child_type}.
The received argument for {child.__class__.__name__.lower()} was {children}.""")


	@restrict
	def _build_children(self, parent, world, indicies, exclude=None) -> None:
		"""
		This methods build the children for xml construction. If some children should 
		be excluded from the build, the names of their attribute can be passed in exclude.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of the childrens parent.
		world : WorldType
			The world from which the build method was called initially.
		exclude : None, optional
			Either a single string or an iterable of strings or the names of children 
			attributes which should be excluded from the build.
		"""
		if exclude is None:
			exclude = ()
		elif isinstance(exclude, str):
			exclude = (exclude,)
		# BUILD CHILDREN
		for name, child_dict in self._CHILDREN.items():
			children = child_dict['children']
			if name not in exclude:
				self._build_tree(children=children, 
						 parent=parent, 
						 world=world, 
						 indicies=indicies)


	@restrict
	def _build_tree(self, 
			children: list[blue.ThingType], 
			parent, 
			world, 
			indicies) -> None:
		"""
		This method build all children from a children attribute.
		
		Parameters
		----------
		children : list[blue.ThingType]
			A list of the children to be build.
		parent : xml.etree.ElementTree.Element
			The xml element of the childrens parent.
		world : WorldType
			The world from which the build method was called initially.
		"""
		for child in children:
			child._build(parent=parent, 
				     world=world, 
				     indicies=indicies)


	def _decouple_descendants(self):
		"""
		This method is used to decouple references of :class:`CyclicalThing` that 
		no longer point inside the kinematic tree. This can be the result of the 
		removal of a Thing. Decoupled references are set to None.
		"""
		for dec_dict in self.descendants.values():
			dec_type = dec_dict['type']
			dec_list = dec_dict['descendants']
			if issubclass(dec_type, blue.CyclicalThingType):
				for dec in dec_list:
					dec._decouple()



import xml.etree.ElementTree as xml
import numpy  as np
import blueprints as blue



class Body(blue.BodyType, blue.thing.MoveableThing, blue.thing.NodeThing):
	"""
	This class is available through the shortcut :class:`blueprints.Body <Body>`.

	Attributes from base classes :class:`MoveableThing <blueprints.thing.moveable.MoveableThing>` and :class:`NodeThing <blueprints.thing.node.NodeThing>` are inherited.
	

A :class:`Body` is the primary Thing used to build the structure of a kinematic tree. This is done by 
attaching them to other bodies as well as other bodies to them. Bodies do have a position and an orientation 
but no further physical properties like a shape, a size or mass. Positions and orientations are local meaning 
that a body is placed relative to its parents frame of reference.To flesh out a :class:`Body` the following 
Things can be attached to it:

* :class:`Geom <blueprints.geoms.BaseGeom>`: This Thing is used to fill in the Body with matter. Geoms have a shape, mass, friction etc.

* :class:`Site <blueprints.sites.BaseSite>`: Sites similar to Geoms have shape, but do not effect the simulation physics. Instead :class:`Sensor <blueprints.sensors.BaseSensor>` instances as well as :class:`Actuator <blueprints.actuators.BaseActuator>` can be attached to it.

* :class:`Camera <blueprints.camera.Camera>`: If attached to the Body, the Camera can be view trough :meth:`World.view <blueprints.world.World.view>` by setting the :attr:`rendering <blueprints.camera.Camera.rendering>` camera value to the Cameras name. If the Body is included in an :class:`Agent <blueprints.agent.Agent>` (not existing right now) the camera will be available in the Agents observations.

* :class:`Placeholder <blueprints.placeholder.Placeholder>`: A placeholder can be used to specify a position and orientation relative to the Body to which it is attached, that is of special interest. If a Thing is attached to a placeholder it is instead attached to its parent in a position and orientation, that matches attachment to the placeholder.

* :class:`Actuator <blueprints.actuators.BaseActuator>`: Actuators are used to apply force to the kinematic tree at the node they are attached to. If the Body is included in an :class:`Agent <blueprints.agent.Agent>` (not yet implemented) the actuators input will appear in the Agents action space.

* :class:`Light <blueprints.light.Light>`: Light sources illuminate the space around them, but are not necessary for the model to be visible, since a base luminescence is presence always present.

* :class:`Joint <blueprints.joints.BaseJoint>`: This Thing is used to move the :class:`Body`:. If a force is applied to the Body it moves along the degrees of freedom defined by the Joint.

.. note::
	Unintuitively Bodies are not connected with each other via a Joint. Instead a Joint attached to a 
	body defines the way in which a Body can be moved w.r.t. to the Bodies parent.

A kinematic tree can either be constructed bottom up or top down. The following two examples are used to 
create a kinematic tree of this structure:

.. code-block:: mxml
	:caption: XML Structure

	<body name="tree">
	    <body pos="0.0 0.0 8.0" name="stem">
	        <body euler="0.0 -0.7853982 0.0" pos="0.0 0.0 4.0" name="branch_(0)">
	            <body euler="0.0 -0.7853982 0.0" pos="0.0 0.0 2.0" name="twig_(0)">
	                <body euler="0.0 -0.7853982 0.0" pos="0.0 0.0 1.0" name="leaf_(0)" />
	                <body euler="0.0 0.7853982 0.0" pos="0.0 0.0 1.0" name="leaf_(1)" />
	            </body>
	            <body euler="0.0 0.7853982 0.0" pos="0.0 0.0 2.0" name="twig_(1)">
	                <body euler="0.0 -0.7853982 0.0" pos="0.0 0.0 1.0" name="leaf_(2)" />
	                <body euler="0.0 0.7853982 0.0" pos="0.0 0.0 1.0" name="leaf_(3)" />
	            </body>
	        </body>
	        <body euler="0.0 0.7853982 0.0" pos="0.0 0.0 4.0" name="branch_(1)">
	            <body euler="0.0 -0.7853982 0.0" pos="0.0 0.0 2.0" name="twig_(2)">
	                <body euler="0.0 -0.7853982 0.0" pos="0.0 0.0 1.0" name="leaf_(4)" />
	                <body euler="0.0 0.7853982 0.0" pos="0.0 0.0 1.0" name="leaf_(5)" />
	            </body>
	            <body euler="0.0 0.7853982 0.0" pos="0.0 0.0 2.0" name="twig_(3)">
	                <body euler="0.0 -0.7853982 0.0" pos="0.0 0.0 1.0" name="leaf_(6)" />
	                <body euler="0.0 0.7853982 0.0" pos="0.0 0.0 1.0" name="leaf_(7)" />
	            </body>
	        </body>
	    </body>
	</body>

.. code-block:: py
	:caption: Bottom-Up

	>>> tree = blue.Body(name='tree')
	>>> stem = blue.Body(name='stem',   pos=[0, 0, 8])
	>>> bran = blue.Body(name='branch', pos=[0, 0, 4])
	>>> twig = blue.Body(name='twig',   pos=[0, 0, 2])
	>>> leaf = blue.Body(name='leaf',   pos=[0, 0, 1])
	>>> angle = TAU/8
	>>> # starting at the leafs
	>>> twig.attach(leaf.rotate(beta=-angle), leaf.rotate(beta=angle))
	>>> bran.attach(twig.rotate(beta=-angle), twig.rotate(beta=angle))
	>>> stem.attach(bran.rotate(beta=-angle), bran.rotate(beta=angle))
	>>> tree.attach(stem)

.. code-block:: py
	:caption: Top-Down

	>>> tree = blue.Body(name='tree')
	>>> stem = blue.Body(name='stem',   pos=[0, 0, 0])
	>>> bran = blue.Body(name='branch', pos=[0, 0, 4])
	>>> twig = blue.Body(name='twig',   pos=[0, 0, 2])
	>>> leaf = blue.Body(name='leaf',   pos=[0, 0, 1])
	>>> angle = TAU/8
	>>> # starting at the stem
	>>> tree.attach(stem)
	>>> tree.bodies.attach(bran.rotate(beta=-angle), bran.rotate(beta=angle))
	>>> tree.bodies.bodies.attach(twig.rotate(beta=-angle), twig.rotate(beta=angle))
	>>> tree.bodies.bodies.bodies.attach(leaf.rotate(beta=-angle), leaf.rotate(beta=angle))


	Attributes
	----------
	bodies, cameras, geoms, sites, joints, lights, cameras, actuators, placeholders : :class:`View <blueprints.utils.view.View>`
		All children that have been attached to the Body are retrieved by the attribute as a :class:`View <blueprints.utils.view.View>`.
	"""
	@blue.restrict
	def __init__(self, 
		     pos:          list[int|float]|np.ndarray = [0., 0., 0.], 
		     alpha:        int|float|None = 0., 
		     beta:         int|float|None = 0., 
		     gamma:        int|float|None = 0., 
		     geoms:        list[blue.ThingType]|blue.ThingType|None = None, 
		     sites:        list[blue.ThingType]|blue.ThingType|None = None, 
		     joints:       list[blue.ThingType]|blue.ThingType|None = None, 
		     bodies:       list[blue.ThingType]|blue.ThingType|None = None, 
		     lights:       list[blue.ThingType]|blue.ThingType|None = None, 
		     cameras:      list[blue.ThingType]|blue.ThingType|None = None, 
		     actuators:    list[blue.ThingType]|blue.ThingType|None = None, 
		     placeholders: list[blue.ThingType]|blue.ThingType|None = None, 
		     name:         str|None       = None, 
		     copy:         bool           = True, 
		     x:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     quat:         list[int|float]|np.ndarray|None = None, 
		     **kwargs) -> None:
		"""
		Parameters
		----------
		pos : list[int | float] | np.ndarray, optional
			Represents the position of the object. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		alpha : int | float | None, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also changes the :attr:`euler` property.
		beta : int | float | None, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also changes the :attr:`euler` property.
		gamma : int | float | None, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also changes the :attr:`euler` property.
		geoms : list[blue.ThingType] | blue.ThingType | None, optional
			The ``geoms`` argument mus either be a list of Geoms or a single Geom.
		sites : list[blue.ThingType] | blue.ThingType | None, optional
			The ``sites`` argument mus either be a list of Sites or a single Site.
		joints : list[blue.ThingType] | blue.ThingType | None, optional
			The ``joints`` argument mus either be a list of Joints or a single Joint.
		bodies : list[blue.ThingType] | blue.ThingType | None, optional
			The ``bodies`` argument mus either be a list of Bodies or a single Body.
		lights : list[blue.ThingType] | blue.ThingType | None, optional
			The ``lights`` argument mus either be a list of Lights or a single Light.
		cameras : list[blue.ThingType] | blue.ThingType | None, optional
			The ``cameras`` argument mus either be a list of Cameras or a single Camera.
		actuators : list[blue.ThingType] | blue.ThingType | None, optional
			The ``actuators`` argument mus either be a list of Actuators or a single Actuator.
		placeholders : list[blue.ThingType] | blue.ThingType | None, optional
			The ``placeholders`` argument mus either be a list of Placeholders or a single Placeholder.
		name : str | None, optional
			The user specified name of the Body. In the case of a naming conflict the name is appended by an enumeration scheme.
		copy : bool, optional
			This argument indicates whether the children should be copied before they are attached. If copy if set to False this will result in graph cycles if a parent of the Body is attached resulting in further errors.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		quat: list [ int | float ] | np.ndarray | None, optional
			If set, the quaternion orientation overwrites the euler angles ``alpha``, ``beta`` and ``gamma``.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		pack = lambda x: x if isinstance(x, list) else [x]
		geoms        = pack(geoms)        if geoms        else []
		joints       = pack(joints)       if joints       else []
		bodies       = pack(bodies)       if bodies       else []
		lights       = pack(lights)       if lights       else []
		sites        = pack(sites)        if sites        else []
		cameras      = pack(cameras)      if cameras      else []
		actuators    = pack(actuators)    if actuators    else []
		placeholders = pack(placeholders) if placeholders else []
		self._geoms             = []
		self._joints            = []
		self._bodies            = []
		self._lights            = []
		self._sites             = []
		self._cameras           = []
		self._actuators         = []
		self._placeholders      = []
		self._targeting_lights  = []
		self._targeting_cameras = []
		self._CHILDREN  = {'lights':            {'type':     blue.LightType, 
							 'children': self._lights}, 
				   'joints':            {'type':     blue.JointType, 
							 'children': self._joints}, 
				   'geoms':             {'type':     blue.GeomType, 
							 'children': self._geoms}, 
				   'sites':             {'type':     blue.SiteType, 
							 'children': self._sites}, 
				   'cameras':           {'type':     blue.CameraType, 
							 'children': self._cameras}, 
				   'actuators':         {'type':     blue.ActuatorType, 
							 'children': self._actuators}, 
				   'placeholders':      {'type':     blue.PlaceholderType, 
							 'children': self._placeholders}, 
				   'bodies':            {'type':     blue.BodyType, 
							 'children': self._bodies}}
		self._CYCLE_REF = {'targeting_lights':  {'type':     blue.CameraType, 
							 'children': self._targeting_lights}, 
				   'targeting_cameras': {'type':     blue.CameraType, 
							 'children': self._targeting_cameras}}
		super().__init__(pos=pos, 
				 x=x, 
				 y=y, 
				 z=z, 
				 alpha=alpha, 
				 beta=beta, 
				 gamma=gamma, 
				 quat=quat, 
				 name=name, 
				 **kwargs)
		self.attach(*geoms, 
			    *joints, 
			    *bodies, 
			    *sites, 
			    *lights, 
			    *cameras, 
			    *actuators, 
			    *placeholders, 
			    copy=copy)
		self._check_children_types()


	@blue.restrict
	def _build(self, parent, world, indicies, **kwargs):
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		indicies : dict
			All indecies used for retrieving data during the simulation.
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Thing.
		"""
		self._index       = indicies['body']
		indicies['body'] += 1
		return super()._build(parent, world, indicies, **kwargs)

	# KINEMATIC TREES PROPERTIES/METHODS

	@blue.restrict
	@classmethod
	def _from_xml_element(cls, 
			      xml_element: xml.Element, 
			      actuators:   list = []) -> blue.ThingType:
		"""
		This method reconstructs a Body from an xml element.
		
		Parameters
		----------
		xml_element : xml.Element
			The xml element from which a Body is reconstructed.
		
		Returns
		-------
		blue.ThingType
			The reconstructed Body.
		"""
		body = super()._from_xml_element(xml_element)
		for actuator in actuators:
			body.attach(actuator, copy=False)
			actuator.body = body
		return body

	# KINEMATIC TREE PROPERTIES

	@property
	def targeting_cameras(self):
		"""
		Targeting Cameras contain :class:`Camera <blueprints.camera.Camera>` that track this Body.
		
		Returns
		-------
		:class:`View <blueprints.utils.view.View>`
		"""
		return blue.View(self._targeting_cameras, name='targeting_cameras', parent=self)


	@targeting_cameras.setter
	@blue.restrict
	def targeting_cameras(self, cameras: blue.CameraType|list[blue.CameraType]|blue.ViewType):
		"""
		Parameters
		----------
		actuators : list[blue.ActuatorType] | blue.ViewType
			Targeting Cameras contain :class:`Camera <blueprints.camera.Camera>` that track this Body.
		"""
		cameras = [cameras] if isinstance(cameras, blue.CameraType) else cameras
		for camera in cameras:
			camera.target = self

	@property
	def targeting_lights(self):
		"""
		Targeting Lights contain :class:`Light <blueprints.light.Light>` that track this Body.
		
		Returns
		-------
		:class:`blueprints.utils.view.View`
		"""
		return blue.View(self._targeting_lights, name='targeting_lights', parent=self)


	@targeting_lights.setter
	@blue.restrict
	def targeting_lights(self, lights: blue.LightType|list[blue.LightType]|blue.ViewType):
		"""
		Parameters
		----------
		actuators : list[blue.ActuatorType] | blue.ViewType
			Targeting Lights contain :class:`Light <blueprints.light.Light>` that track this Body.
		"""
		lights = [lights] if isinstance(lights, blue.LightType) else lights
		for light in lights:
			light.target = self
